#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **application metadata singleton** (i.e., application-wide object
synopsizing application metadata via read-only properties).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: This subclass must *NOT* be accessed from the top-level "setup.py"
# script of this or any other application. This application and hence this
# subclass is *NOT* guaranteed to exist at setuptools-based installation-time
# for downstream consumers (e.g., BETSEE).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import betse
from betse import metadata, metadeps
from betse.util.app import apppath
from betse.util.app.meta.appmetaabc import AppMetaABC
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.types import type_check, ModuleType

# ....................{ SUBCLASSES                         }....................
class BetseAppMeta(AppMetaABC):
    '''
    **Application metadata singleton** (i.e., application-wide object
    synopsizing application metadata via read-only properties) subclass.

    Caveats
    ----------
    **This subclass must not be accessed from the top-level ``setup.py`` script
    of this or any other application.** This application and hence this
    subclass is *not* guaranteed to exist at setuptools-based installation-time
    for downstream consumers (e.g., BETSEE).
    '''

    # ..................{ SUPERCLASS ~ properties            }..................
    @property
    def _module_metadata(self) -> ModuleType:
        return metadata

    #FIXME: *THIS SHOULD NO LONGER BE CALLED ANYWHERE.* Excise, please.
    @property
    def _module_metadeps(self) -> ModuleType:
        return metadeps

    # ..................{ PROPERTIES ~ dir                   }..................
    @property_cached
    def betse_data_dirname(self) -> str:
        '''
        Absolute dirname of BETSE's top-level data directory if found *or*
        raise an exception otherwise (i.e., if this directory is *not* found).

        This directory typically contains BETSE-internal resources (e.g., media
        files) required at BETSE runtime.

        Design
        ----------
        The dirname returned by this property is guaranteed to be identical to
        the dirname returned by the :meth:`data_dirname` *only* if the current
        application is BETSE. If the current application is instead a
        downstream consumer of BETSE (e.g., BETSEE, PLIMBO), the dirname
        returned by this property is guaranteed to be BETSE-specific while that
        returned by the :meth:`data_dirname` property is guaranteed to be
        downstream-specific and hence differ from the former.

        In short, this property enables downstream consumers to access
        BETSE-specific data in a non-ambiguous manner *without* inviting
        inheritance or subclass issues.

        Raises
        ----------
        BetseDirException
            If this directory does *not* exist.
        '''

        # Return the absolute dirname of this BETSE-relative directory if
        # this directory exists *OR* raise an exception otherwise.
        return apppath.get_dirname(package=betse, dirname='data')


    @property_cached
    def betse_data_yaml_dirname(self) -> str:
        '''
        Absolute pathname of this application's data subdirectory containing
        YAML-formatted files if found *or* raise an exception otherwise (i.e.,
        if this directory is *not* found).

        This directory contains:

        * The YAML-formatted plaintext file specifying a simulation
          configuration.
        * All assets referenced in and hence required by this file.

        Raises
        ----------
        BetseDirException
            If this directory does *not* exist.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import dirs

        # Return this dirname if this directory exists or raise an exception.
        return dirs.join_or_die(self.betse_data_dirname, 'yaml')

    # ..................{ PROPERTIES ~ file                  }..................
    @property_cached
    def betse_sim_conf_default_filename(self) -> str:
        '''
        Absolute pathname of this application's **default simulation
        configuration file** (i.e., plaintext YAML-formatted file from which
        all simulation configurations derive) if found *or* raise an exception
        otherwise (i.e., if this file is *not* found).

        Raises
        ----------
        BetseFileException
            If this file does *not* exist.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import files

        # Return this dirname if this directory exists or raise an exception.
        return files.join_or_die(
            self.betse_data_yaml_dirname, 'sim_config.yaml')

    # ..................{ GETTERS                            }..................
    @type_check
    def get_repl_history_filename(self, repl_module_name: str) -> dict:
        '''
        Absolute pathname of this application's default user-specific history
        file for the read–eval–print loop (REPL) implemented by the third-party
        module with the passed name.

        This history file is a plaintext file to which this REPL appends each
        read user command for preserving command history between REPL sessions.

        Parameters
        ----------
        repl_module_name : str
            Fully-qualified name of the third-party module implementing this
            REPL.

        Returns
        ----------
        str
            Absolute path of the user-specific history file for this REPL.

        Raises
        ----------
        BetseModuleException
            If this REPL module name is unrecognized.
        '''

        # Avoid circular import dependencies.
        from betse.exceptions import BetseModuleException
        from betse.util.path import pathnames

        # Dictionary mapping each REPL module name to its history filename.
        REPL_MODULE_NAME_TO_HISTORY_FILENAME = {
            'ptpython': pathnames.join(self.dot_dirname, 'ptpython.hist'),
            'readline': pathnames.join(self.dot_dirname, 'readline.hist'),
        }

        # If this REPL module name is unrecognized, fail.
        if repl_module_name not in REPL_MODULE_NAME_TO_HISTORY_FILENAME:
            raise BetseModuleException(
                'REPL module "{}" unrecognized.'.format(repl_module_name))

        # Else, return the history filename for this REPL.
        return REPL_MODULE_NAME_TO_HISTORY_FILENAME[repl_module_name]



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **enumeration** (i.e., enumerable types created by the `Enum()` class)
facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseEnumException
from betse.util.type.types import (
    type_check,
    CallableTypes,
    EnumMemberType,
    EnumType,
    EnumClassType,
    GeneratorType,
    SequenceTypes,
    StrOrNoneTypes,
)
from functools import partial

# ....................{ SUBCLASSES                        }....................
class EnumOrdered(EnumClassType):
    '''
    Enumeration whose members are comparable according to their assigned
    values.

    This :class:`Enum` subclass complies with the `Functional API`_ for
    enumerations, permitting comparable enumeration types to be defined with a
    single function call. (See the example below.)

    Oddly, the :class:`Enum` superclass does *not* support such comparisons.
    This :class:`Enum` subclass amends this oversight, implementing four of the
    six rich comparison special methods. The fifth and sixth (i.e.,
    ``__eq__()`` and ``__ne__()``) are already implemented by the :class:`Enum`
    superclass and hence need *not* be reimplemented here.

    .. _Functional API:
       https://docs.python.org/3/library/enum.html#functional-api

    See Also
    ----------
    https://docs.python.org/3/library/enum.html#orderedenum
        Standard Python documentation strongly inspiring this implementation.

    Examples
    ----------
        >>> from betse.util.type.enums import EnumOrdered
        >>> Shrike = EnumOrdered('Shrike', ('hyperion', 'endymion'))
        >>> Shrike.endymion == Shrike.endymion
        True
        >>> Shrike.hyperion != Shrike.endymion
        True
        >>> Shrike.hyperion < Shrike.endymion
        True
        >>> Shrike.hyperion <= Shrike.endymion
        True
        >>> Shrike.endymion > Shrike.hyperion
        True
        >>> Shrike.endymion >= Shrike.hyperion
        True
    '''

    # ..................{ COMPARATORS                       }..................
    # For efficiency, rich comparison special methods are expected to return
    # the "NotImplemented" constant rather than raise the "NotImplementedError"
    # exception when the current and passed objects are incomparible.

    def __ge__(self, other: object) -> bool:
        '''
        :data:`NotImplemented` if the passed object is *not* a member of the
        same enumeration as this member, ``True`` if the value of this member
        is greater than or equal to that of the passed member, or ``False``
        otherwise.
        '''

        return (
            self.value >= other.value if self.__class__ is other.__class__ else
            NotImplemented)


    def __gt__(self, other: object) -> bool:
        '''
        :data:`NotImplemented` if the passed object is *not* a member of the
        same enumeration as this member, ``True`` if the value of this member
        is greater than that of the passed member, or ``False`` otherwise.
        '''

        return (
            self.value > other.value if self.__class__ is other.__class__ else
            NotImplemented)


    def __le__(self, other: object) -> bool:
        '''
        :data:`NotImplemented` if the passed object is *not* a member of the
        same enumeration as this member, ``True`` if the value of this member
        is less than or equal to that of the passed member, or ``False``
        otherwise.
        '''

        return (
            self.value <= other.value if self.__class__ is other.__class__ else
            NotImplemented)


    def __lt__(self, other: object) -> bool:
        '''
        :data:`NotImplemented` if the passed object is *not* a member of the
        same enumeration as this member, ``True`` if the value of this member
        is less than that of the passed member, or ``False`` otherwise.
        '''

        return (
            self.value < other.value if self.__class__ is other.__class__ else
            NotImplemented)

# ....................{ MAKERS                            }....................
@type_check
def make_enum(
    # Mandatory parameters.
    class_name: str,
    member_names: SequenceTypes,

    # Optional parameters.
    is_ordered: bool = False,
    doc: StrOrNoneTypes = None,
) -> EnumType:
    '''
    **Integer-based enumeration type** (i.e., instance of the standard
    :class:`EnumClassType` type whose members are uniquely mapped to 1-based
    integers) dynamically synthesized to have the passed class name and contain
    only the members with the passed names.

    This factory function is a convenience wrapper for the
    :class:`EnumClassType` subclass, whose non-standard semantics are arguably
    more obfuscatory than helpful.

    Attributes
    ----------
    Each attribute of the returned enumeration type is an enumeration member
    whose:

    * Type is the same as that of this enumeration.
    * Value is an object uniquely identifying this member in this enumeration,
      defining the following public attributes:

      * ``name``, the Python identifier uniquely identifying this member --
        guaranteed to be the same string as that passed to this function.
      * ``value``, the 1-based integer uniquely identifying this member. The
        ``value`` of this enumeration's:

        * First member is guaranteed to be 1.
        * Last member is guaranteed to be the number of members (i.e., the
          length of this enumeration).

    Parameters
    ----------
    class_name : str
        Class name of this enumeration type, ideally unique across all
        attributes of the module calling this function.
    member_names : SequenceTypes
        Sequence of the names of all members of this enumeration type, required
        to be valid **Python identifiers** (i.e., contain only alphanumeric
        characters and underscores).
    is_ordered : bool
        ``True`` only if the members of this enumeration are ordered (and hence
        comparable) according to the ordering of member names in the passed
        ``member_names`` sequence. Defaults to ``False``, in which case the
        members of this enumeration are unordered (and hence incomparable).
    doc : StrOrNoneTypes
        Class docstring to document this type with. Defaults to ``None``, in
        which case this type remains undocumented.

    Returns
    ----------
    EnumType
        Enumeration type dynamically synthesized as defined as above.

    Examples
    ----------
        >>> from betse.util.type.enums import make_enum
        >>> CthulhicState = make_enum(
        ...     class_name='CthulhicState',
        ...     member_names=('MERCIFUL', 'INNABILITY', 'CORRELATE', 'CONTENTS',))
        >>> CthulhicState.CORRELATE.name
        'CORRELATE'
        >>> CthulhicState.CORRELATE.value
        3
    '''

    # Avoid circular import dependencies.
    from betse.util.py import pyident
    from betse.util.type.call import callers
    from betse.util.type.iterable import itertest

    # If the passed class name is syntactically invalid, raise an exception.
    pyident.die_unless_unqualified(class_name)

    # If any passed member name is *NOT* a string, raise an exception.
    itertest.die_unless_items_instance_of(iterable=member_names, cls=str)

    # Fully-qualified name of the module defining this enumeration type.
    # Note that the obfuscatory magic "4" here is due to the additional
    # stack frames implicitly added to the stack by the @type_check
    # decorator of both this function *AND* the get_caller_module_name()
    # function. Without those frames, this would simply be "2" as expected.
    module_name = callers.get_caller_module_name(call_stack_index=4)
    # print(f'make_enum(): module_name={module_name}')

    # Type of enumeration superclass to be subclassed. Specifically, if the
    # caller requested that the members of this enumeration be:
    #
    # * Ordered, this is our custom ordered enumeration type.
    # * Unordered, this is the standard unordered enumeration type.
    enum_superclass = EnumOrdered if is_ordered else EnumClassType

    # Dynamically synthesize and return this enumeration type.
    enum_subclass = enum_superclass(
        value=class_name,
        names=member_names,
        module=module_name,
        qualname=class_name,
#        qualname='{}.{}'.format(module_name, class_name),
    )

    # If passed a docstring, assign this subclass this docstring.
    if doc is not None:
        enum_subclass.__doc__ = doc
#    print(f'enum_sublass={enum_subclass.__module__}, {enum_subclass.__name__}, {enum_subclass.__qualname__}')

    # Return this subclass.
    return enum_subclass

# ....................{ EXCEPTIONS                        }....................
def die_unless_member(
    enum_type: EnumType, enum_member: EnumMemberType) -> None:
    '''
    Raise an exception unless the passed enumeration contains the passed
    member.

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.
    enum_member: str
        Enumeration members to test for.
    '''

    if not is_member(enum_type, enum_member):
        raise BetseEnumException(
            'Enumeration {} member "{}" not found.'.format(
                enum_type.__name__, enum_member.name))


def die_unless_member_name(
    enum_type: EnumType, enum_member_name: str) -> None:
    '''
    Raise an exception unless the passed enumeration contains an enumeration
    member with the passed name.

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.
    enum_member_name: str
        Name of the member to test for.
    '''

    if not is_member_name(enum_type, enum_member_name):
        raise BetseEnumException(
            'Enumeration {} member "{}" not found.'.format(
                enum_type.__name__, enum_member_name))

# ....................{ TESTERS                           }....................
@type_check
def is_member(enum_type: EnumType, enum_member: EnumMemberType) -> bool:
    '''
    ``True`` only if the passed enumeration contains the passed member.

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.
    enum_member_name: str
        Name of the member to test for.

    Returns
    ----------
    bool
        ``True`` only if this enumeration contains this member.
    '''

    # No, thank *YOU*, Enum.__contains__().
    return enum_member in enum_type


@type_check
def is_member_name(enum_type: EnumType, enum_member_name: str) -> bool:
    '''
    ``True`` only if the passed enumeration contains an enumeration member with
    the passed name.

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.
    enum_member_name: str
        Name of the member to test for.

    Returns
    ----------
    bool
        ``True`` only if this enumeration contains a member with this name.

    See Also
    ----------
    https://stackoverflow.com/a/29795561/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # This is insanity, but insanity that works. (Yes, this is the accepted
    # method for testing enumeration member existence by name. Thanks, Guido.)
    return enum_member_name in enum_type.__members__

# ....................{ GETTERS ~ enum                    }....................
@type_check
def get_converter_name_to_uppercase_enum_member(
    enum_type: EnumType) -> CallableTypes:
    '''
    Callable accepting one string parameter returning the member of the passed
    enumeration whose name is that string uppercased, raising an exception if
    this string uppercased is *not* the name of a member of this enumeration.

    The callable returned by this function is principally intended to be passed
    as the ``type`` parameter to the :meth:`ArgumentParser.add_argument`
    method, converting from lowercase command-line option string arguments to
    corresponding enumeration members.

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.

    Returns
    ----------
    CallableTypes
        Callable whose signature is defined as above.
    '''

    # Return a closure passing the passed enumeration to an existing getter.
    return partial(get_member_from_name_uppercased, enum_type=enum_type)


@type_check
def get_member_names_lowercase(enum_type: EnumType) -> SequenceTypes:
    '''
    Sequence of the unqualified lowercased names of all members of the passed
    enumeration type in **declaration order** (i.e., the order in which these
    members were originally declared).

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.

    Returns
    ----------
    SequenceTypes
        Sequence of the lowercased names of all members of this enumeration.
    '''

    return tuple(iter_names_lowercase(enum_type))

# ....................{ GETTERS ~ member                  }....................
@type_check
def get_enum_name_from_member(enum_member: EnumMemberType) -> str:
    '''
    Unqualified name of the enumeration class to which the passed enumeration
    member belongs (e.g., ``SolverType`` given a member ``SolverType.FULL``).

    Parameters
    ----------
    enum_member : EnumMemberType
        Enumeration member to be inspected.

    Returns
    ----------
    str
        Unqualified name of this member's enumeration class.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # Strange, but true. We don't ask questions. Neither should you.
    return objects.get_class_name_unqualified(enum_member)


@type_check
def get_class_member_name(enum_member: EnumMemberType) -> str:
    '''
    Partially-qualified name of the passed enumeration member (e.g.,
    ``SolverType.FULL``), excluding the module defining this enumeration.

    Parameters
    ----------
    enum_member : EnumMemberType
        Enumeration member to be inspected.

    Returns
    ----------
    str
        Partially-qualified name of this member.

    See Also
    ----------
    :func:`get_member_name`
        Unqualified name of this member.
    '''

    # Makes sense.
    return str(enum_member)


@type_check
def get_member_name(enum_member: EnumMemberType) -> str:
    '''
    Unqualified name of the passed enumeration member (e.g., ``FULL`` given a
    member ``SolverType.FULL``).

    Since the :attr:`enum_member.name` attribute already provides this name,
    this function principally exists for orthogonality with other getters.

    Parameters
    ----------
    enum_member : EnumMemberType
        Enumeration member to be inspected.

    Returns
    ----------
    str
        Unqualified name of this member.

    See Also
    ----------
    :func:`get_class_member_name`
        Partially-qualified name of this member.
    '''

    # Believe it.
    return enum_member.name


@type_check
def get_member_name_lowercase(enum_member: EnumMemberType) -> str:
    '''
    Unqualified lowercase name of the passed enumeration member (e.g., ``full``
    given a member ``SolverType.FULL``).

    Parameters
    ----------
    enum_member : EnumMemberType
        Enumeration member to be inspected.

    Returns
    ----------
    str
        Unqualified lowercase name of this member.
    '''

    # It cannot be, yet it is.
    return enum_member.name.lower()

# ....................{ GETTERS ~ member : name           }....................
@type_check
def get_member_from_name(
    enum_type: EnumType, enum_member_name: str) -> EnumMemberType:
    '''
    Member of this enumeration with the passed name if this member exists *or*
    raise an exception otherwise (i.e., if this member does *not* exist).

    Parameters
    ----------
    enum_type: EnumType
        Enumeration type to inspect.
    enum_member_name : EnumType
        Name of the member of this enumeration to be returned.

    Returns
    ----------
    EnumMemberType
        Member of this enumeration with this name.

    Raises
    ----------
    BetseEnumException
        If no member of this enumeration with this name exists.
    '''

    # If no member with this name exists, raise an exception.
    die_unless_member_name(enum_type, enum_member_name)

    # Return this member.
    return enum_type[enum_member_name]


@type_check
def get_member_from_name_uppercased(
    enum_type: EnumType, enum_member_name: str) -> EnumMemberType:
    '''
    Member of this enumeration with the passed name uppercased if this member
    exists *or* raise an exception otherwise (i.e., if this member does *not*
    exist).

    Parameters
    ----------
    enum_type: EnumType
        Enumeration type to inspect.
    enum_member_name : EnumType
        String that when uppercased yields the name of the member of this
        enumeration to be returned.

    Returns
    ----------
    EnumMemberType
        Member of this enumeration with this name uppercased.

    Raises
    ----------
    BetseEnumException
        If no member of this enumeration with this name uppercased exists.
    '''

    # It pays to be lazy. Admittedly, it doesn't pay much.
    return get_member_from_name(
        enum_type=enum_type, enum_member_name=enum_member_name.upper())

# ....................{ GETTERS ~ member : value          }....................
@type_check
def get_member_from_value(
    enum_type: EnumType, enum_member_value: object) -> EnumMemberType:
    '''
    Member of this enumeration associated with the passed arbitrary value if
    this member exists *or* raise an exception otherwise (i.e., if this member
    does *not* exist).

    Parameters
    ----------
    enum_type: EnumType
        Enumeration type to inspect.
    enum_member_value : object
        Arbitrary value associated with the member of this enumeration to be
        returned.

    Returns
    ----------
    EnumMemberType
        Member of this enumeration associated with this value.

    Raises
    ----------
    ValueError
        If no member of this enumeration with this value exists.
    '''

    # Return this member. Yes, for some obscure, inane, and unintelligible
    # reason, the "Enum" superclass performs name lookup by indexing syntax and
    # value lookup by... instantiating the "Enum" subclass. It makes utterly no
    # sense, but there it is. (Rationality has left the building.)
    #
    # Note that enumerations provide efficient mechanisms for determining
    # whether or not a member with a given name but *NOT* value exists. See the
    # standard enum.Enum.__new__() function for the supporting evidence. Ergo,
    # there intentionally exists no die_unless_member_value() function; while
    # such a function could be defined, doing so would effectively reduce to:
    #
    #     try:
    #         enum_type(enum_member_value)
    #     except ValueError:
    #         raise BetseEnumException(
    #             'Enumeration {} member with value {!r} not found.'.format(
    #                 enum_type.__name__, enum_member_value))
    #
    # To avoid such uselessly repetitious invocations of the enum_type()
    # pseudo-method, we simply call this pseudo-method as is and defer to the
    # mostly human-readable exception it already raises.
    return enum_type(enum_member_value)

# ....................{ ITERATORS                         }....................
@type_check
def iter_names(enum_type: EnumType) -> GeneratorType:
    '''
    Generator yielding the unmodified name of each member of the passed
    enumeration type in **declaration order** (i.e., the order in which these
    members were originally declared).

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.

    Returns
    ----------
    GeneratorType
        Generator yielding the unmodified name of each member of this
        enumeration type.
    '''

    return (enum_member.name for enum_member in enum_type)


@type_check
def iter_names_lowercase(enum_type: EnumType) -> GeneratorType:
    '''
    Generator yielding the lowercased name of each member of the passed
    enumeration type in **declaration order** (i.e., the order in which these
    members were originally declared).

    Parameters
    ----------
    enum_type : EnumType
        Enumeration type to be inspected.

    Returns
    ----------
    GeneratorType
        Generator yielding the lowercased name of each member of this
        enumeration type.
    '''

    return (enum_member.name.lower() for enum_member in enum_type)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Visual export subconfiguration** (i.e., YAML-backed wrapper exposing settings
globally applicable to all visual exports) functionality.
'''

#FIXME: Refactor all remaining visual export-centric settings currently defined
#as "Parameters" instance variables into data descriptors defined below.

# ....................{ IMPORTS                           }....................
from betse.lib.yaml.yamlalias import yaml_alias
from betse.lib.yaml.abc.yamlabc import YamlABC
# from betse.util.type.types import type_check

# ....................{ SUBCLASSES                        }....................
class SimConfExportVisual(YamlABC):
    '''
    **Visual export subconfiguration** (i.e., YAML-backed wrapper exposing
    settings globally applicable to all visual exports).

    Attributes (Cell: Indices)
    ----------
    is_show_cell_indices : bool
        ``True`` only if the 0-based indices of all cells are to be displayed
        over all cell cluster visuals (e.g., as integers situated at cell
        centres).
    single_cell_index : int
        0-based index of the cell to be visualized for all single cell visuals.
        Defaults to 0, the index assigned to the first cell guaranteed to
        exist. Note that cell indices are seed-specific and may be visualized
        by enabling the :attr:`is_show_cell_indices` boolean.
    '''

    # ..................{ ALIASES ~ cell : index            }..................
    is_show_cell_indices = yaml_alias(
        "['results options']['visuals']['cell indices']['show']", bool)

    #FIXME, Consider generalizing this into a list.
    single_cell_index = yaml_alias(
        "['results options']['visuals']['cell indices']['single cell']", int)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level Python facilities pertaining to the active Python interpreter.

Caveats
-------
Word size-specific functions (e.g., :func:`.is_wordsize_64`) are generally
considered poor form. Call these functions *only* where necessary.
'''

# ....................{ IMPORTS                            }....................
import platform, sys
from betse import metadata
from betse.exceptions import BetsePyException
from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import (
    type_check, MappingOrNoneTypes, SequenceTypes)

# ....................{ INITIALIZERS                       }....................
def init() -> None:
    '''
    Validate the active Python interpreter.

    This function does *not* validate this interpreter's version, as the
    top-level :mod:`betse.metadata` submodule already does so at the start of
    application startup. Instead, this function (in order):

    #. Logs a non-fatal warning if this interpreter is *not* 64-bit.
    '''

    # Log this validation.
    logs.log_debug('Validating Python interpreter...')

    # If this Python interpreter is 32- rather than 64-bit, log a non-fatal
    # warning. While technically feasible, running BETSE under 32-bit Python
    # interpreters imposes non-trivial constraints detrimental to sanity.
    if is_wordsize_32():
        logs.log_warning(
            '32-bit Python interpreter detected. '
            '{name} will be confined to low-precision datatypes and '
            'at most 4GB of available RAM, '
            'impeding the reliability and scalability of modelling. '
            'Consider running {name} only under a '
            '64-bit Python interpreter.'.format(name=metadata.NAME))

# ....................{ TESTERS                            }....................
@func_cached
def is_conda() -> bool:
    '''
    ``True`` only if the active Python interpreter is managed by ``conda``, the
    open-source, cross-platform, language-agnostic package manager provided by
    the Anaconda and Miniconda distributions.

    Specifically, this function returns ``True`` only if the
    ``{sys.prefix}/conda-meta/history`` file exists such that:

    * ``{sys.prefix}`` is the site-specific Python directory prefix. Under
      Linux, this is typically:

      * ``/usr`` for system-wide Python interpreters.
      * A user-specific directory for ``conda``-based Python interpreters
        (e.g., ``${HOME}/Miniconda3/envs/${CONDA_ENV_NAME}``).

    * ``conda-meta/history`` is a file relative to this prefix guaranteed to
      exist for *all* ``conda`` environments (including the base environment).
      Since this file is unlikely to be created in non-``conda`` environments,
      the existence of this file effectively guarantees this interpreter to be
      managed by ``conda``.

    See Also
    ----------
    https://stackoverflow.com/a/47730405/2809027
        StackOverflow answer strongly inspiring this implementation.
    https://github.com/conda/constructor/blob/2.0.1/constructor/install.py#L218-L234
        Function in the ``conda`` codebase responsible for creating the
        ``{sys.prefix}/conda-meta/history`` file.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import files, pathnames

    # Absolute filename of a file relative to the site-specific Python
    # directory prefix guaranteed to exist for all "conda" environments.
    conda_history_filename = pathnames.join(
        sys.prefix, 'conda-meta', 'history')

    # Return true only if this file exists.
    return files.is_file(conda_history_filename)

# ....................{ TESTERS ~ wordsize                 }....................
@func_cached
def is_wordsize_32() -> bool:
    '''
    ``True`` only if the active Python interpreter is **32-bit** (i.e., was
    compiled with a 32-bit toolchain into a 32-bit executable).
    '''

    return not is_wordsize_64()


@func_cached
def is_wordsize_64() -> bool:
    '''
    ``True`` only if the active Python interpreter is **64-bit** (i.e., was
    compiled with a 64-bit toolchain into a 64-bit executable).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.numeric import ints

    # Return whether or not the maximum integer size supported by this Python
    # interpreter is larger than the maximum value for variables of internal
    # type `Py_ssize_t` under 32-bit Python interpreters. While somewhat obtuse,
    # this condition is well-recognized by the Python community as the optimal
    # means of portably testing this. Less preferable alternatives include:
    #
    # * "return 'PROCESSOR_ARCHITEW6432' in os.environ", which depends upon
    #   optional environment variables and hence is clearly unreliable.
    # * "return platform.architecture()[0] == '64bit'", which fails under:
    #   * OS X, returning "64bit" even when the active Python interpreter is a
    #     32-bit executable binary embedded in a so-called "universal binary."
    return sys.maxsize > ints.INT_VALUE_MAX_32_BIT

# ....................{ GETTERS                            }....................
@func_cached
def get_wordsize() -> int:
    '''
    Size in bits of variables of internal type `Py_ssize_t` for the active
    Python interpreter.

    Returns
    ----------
    int
        If the active Python interpreter is:

        * 64-bit, 64.
        * 32-bit, 64.
    '''

    return 64 if is_wordsize_64() else 32


def get_version() -> str:
    '''
    Human-readable ``.``-delimited string specifying the most recent version of
    the Python language supported by the active Python interpreter (e.g.,
    ``2.7.10``, ``3.4.1``).
    '''

    return platform.python_version()

# ....................{ GETTERS ~ path                     }....................
@func_cached
def get_command_line_prefix() -> list:
    '''
    List of one or more shell words unambiguously running the executable binary
    for the active Python interpreter and machine architecture.

    Since the absolute path of the executable binary for the active Python
    interpreter is insufficient to unambiguously run this binary under the
    active machine architecture, this function should *always* be called in
    lieu of :func:`get_filename` when attempting to rerun this interpreter as a
    subprocess of the current Python process. As example:

    * Under macOS, the executable binary for this interpreter may be bundled
      with one or more other executable binaries targetting different machine
      architectures (e.g., 32-bit, 64-bit) in a single so-called "universal
      binary." Distinguishing between these bundled binaries requires passing
      this interpreter to a prefixing macOS-specific command, ``arch``.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import macos

    # List of such shell words.
    command_line = None

    # If this is OS X, this interpreter is only unambiguously runnable via the
    # OS X-specific "arch" command.
    if macos.is_macos():
        # Run the "arch" command.
        command_line = ['arch']

        # Instruct this command to run the architecture-specific binary in
        # Python's universal binary corresponding to the current architecture.
        if is_wordsize_64():
            command_line.append('-i386')
        else:
            command_line.append('-x86_64')

        # Instruct this command, lastly, to run this interpreter.
        command_line.append(get_filename())
    # Else, this interpreter is unambiguously runnable as is.
    else:
        command_line = [get_filename()]

    # Return this list.
    return command_line


@func_cached
def get_filename() -> str:
    '''
    Absolute filename of the executable binary for the active Python process.
    '''

    # Absolute filename of this executable if this path is retrievable by
    # Python *OR* either "None" or the empty string otherwise.
    py_filename = sys.executable

    # If this filename is *NOT* retrievable, raise an exception.
    if not py_filename:
        raise BetsePyException(
            'Absolute filename of Python interpreter not retrievable.')

    # Return this filename.
    return py_filename


@func_cached
def get_shebang() -> str:
    '''
    POSIX-compliant **shebang** (i.e., machine-readable ``#!``-prefixed first
    line of interpretable scripts) unambiguously running the executable binary
    for the active Python interpreter and machine architecture.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    # List of one or more shell words unambiguously running this binary.
    py_command_args = get_command_line_prefix()

    # Space-delimited string listing these words.
    py_command = strjoin.join_on_space(py_command_args)

    # Return this string prefixed by the shebang identifier.
    return '#!' + py_command

# ....................{ GETTERS ~ metadata                 }....................
def get_metadata() -> 'betse.util.type.iterable.mapping.mapcls.OrderedArgsDict':
    '''
    Ordered dictionary synopsizing the active Python interpreter.
    '''

    # Avoid circular import dependencies.
    from betse.util.py import pyfreeze
    from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict

    # Return this dictionary.
    return OrderedArgsDict(
        'version',  get_version(),
        'wordsize', get_wordsize(),
        'conda',    is_conda(),
        'frozen',   pyfreeze.is_frozen(),
    )

# ....................{ RUNNERS                            }....................
@type_check
def rerun_or_die(
    command_args: SequenceTypes,
    popen_kwargs: MappingOrNoneTypes = None) -> None:
    '''
    Rerun the active Python interpreter as a subprocess of the current Python
    process, raising an exception on subprocess failure.

    Parameters
    ----------
    command_args : SequenceTypes
        List of zero or more arguments to pass to this interpreter.
    popen_kwargs : optional[MappingType]
        Dictionary of all keyword arguments to pass to the
        :meth:`subprocess.Popen.__init__` method. Defaults to :data:`None`, in
        which case the empty dictionary is assumed.

    See Also
    --------
    :func:`betse.util.os.command.cmdrun.run_or_die`
        Low-level commentary on subprocess execution.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.command import cmdrun

    # List of one or more shell words comprising this command.
    command_words = get_command_line_prefix() + command_args

    # Rerun this interpreter.
    return cmdrun.run_or_die(
        command_words=command_words, popen_kwargs=popen_kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level temporary path facilities, including both temporary directories _and_
non-directory files.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.types import type_check, StrOrNoneTypes
from tempfile import NamedTemporaryFile

# ....................{ WRITERS                            }....................
# The type of the return value is the private class
# "tempfile._TemporaryFileWrapper", which due to being private is intentionally
# *NOT* type-checked here.
@type_check
def write_bytes(encoding: StrOrNoneTypes = None):
    '''
    Open and return a temporary named binary file for byte-oriented writing.

    Parameters
    ----------
    encoding : optional[str]
        Name of the encoding with which to encode bytes into plaintext strings
        if desired or `None` otherwise (i.e., if bytes are to be written as is
        without being encoded into such strings). Defaults to `None`.

    Returns
    ----------
    tempfile._TemporaryFileWrapper
        `file`-like object encapsulating the opened file.

    See Also
    ----------
    :func:`writing_chars`
        Further details.
    '''

    return NamedTemporaryFile(delete=False, encoding=encoding)


# The type of the return value is the private class
# "tempfile._TemporaryFileWrapper", which due to being private is intentionally
# *NOT* type-checked here.
@type_check
def write_chars(encoding: str = 'utf-8'):
    '''
    Open and return a temporary named plaintext file for line-oriented writing.

    This function returns a `file`-like object, suitable for use wherever the
    builtin `open()` would otherwise be called (e.g., in `with` statements).
    The filename for the underlying file is accessible in the typical way for
    named `file`-like objects (namely, via the `name` attribute of such object).
    The underlying file will _not_ be implicitly deleted when such object is
    closed, which typically defeats the purpose of creating a temporary named
    file in the first place.

    Text Lines
    ----------
    Temporary files are _not_ reliably openable for line-oriented writing in a
    cross-platform manner. Hence, callers are discouraged from calling this
    function except where absolutely required (e.g., some or all underlying
    encodings are unknown). Instead, text lines may be manually written in a
    byte-oriented manner by encoding such lines under a known encoding _and_
    appending such lines by the newline delimiter specific to the current
    platform: e.g.,

        >>> from betse.util.path imports temps
        >>> import os
        >>> tempfile = temps.writing_bytes(encoding='utf-8')
        >>> tempfile.write(b'The strength of a civilization is not measured ')
        >>> tempfile.write(b'by its ability to fight wars, but rather by its ')
        >>> tempfile.write(b'ability to prevent them.')
        >>> tempfile.write(bytes(os.linesep, encoding='utf-8'))

    Parameters
    ----------
    encoding : optional[str]
        Name of the encoding to be used. Defaults to UTF-8.

    Returns
    ----------
    tempfile._TemporaryFileWrapper
        `file`-like object encapsulating the opened file.
    '''

    return NamedTemporaryFile(mode='w+', delete=False, encoding=encoding)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **mapping tester** (i.e., utility functions testing and validating
dictionary-like types or instances) functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import (
    BetseMappingException,
    BetseMappingKeyException,
    BetseMappingValueException,
)
from betse.util.type.types import (
    type_check, IterableTypes, MappingType, HashableType)

# ....................{ EXCEPTIONS ~ map : keys           }....................
@type_check
def die_unless_has_keys(mapping: MappingType, keys: IterableTypes) -> None:
    '''
    Raise an exception unless the passed dictionary contains *all* passed keys.

    Equivalently, this function raises an exception if this dictionary does
    *not* contain one or more passed keys.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be validated.
    keys : IterableTypes
        Iterable of all keys to be tested for.

    Raises
    ----------
    BetseMappingKeyException
        If this dictionary does *not* contain one or more passed keys.

    See Also
    ----------
    :func:`has_keys`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    # If this dictionary does *NOT* contain one or more passed keys...
    if not has_keys(mapping=mapping, keys=keys):
        # Set of all passed keys *NOT* in this dictionary.
        keys_missing = set(key for key in keys if key not in mapping)

        # Grammatically proper noun describing the number of such keys.
        keys_noun = 'key' if len(keys_missing) == 1 else 'keys'

        # Raise an exception embedding this set.
        raise BetseMappingKeyException(
            'Dictionary {} {} not found.'.format(
                keys_noun,
                strjoin.join_as_conjunction_double_quoted(*keys_missing)))

# ....................{ EXCEPTIONS ~ map : values         }....................
@type_check
def die_unless_values_unique(mapping: MappingType) -> None:
    '''
    Raise an exception unless all values of the passed dictionary are unique.

    Equivalently, this function raises an exception if any two key-value pairs
    of this dictionary share the same values.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be validated.

    Raises
    ----------
    BetseMappingValueException
        If at least one value of this dictionary is a duplicate.

    See Also
    ----------
    :func:`is_values_unique`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import iterget
    from betse.util.type.text.string import strjoin

    # If one or more values of this dictionary are duplicates...
    if not is_values_unique(mapping):
        # Set of all duplicate values in this dictionary.
        values_duplicate = iterget.get_items_duplicate(mapping.values())

        # Grammatically proper noun describing the number of such values.
        values_noun = 'value' if len(values_duplicate) == 1 else 'values'

        # Raise an exception embedding this set.
        raise BetseMappingValueException(
            'Dictionary {} {} duplicate.'.format(
                values_noun,
                strjoin.join_as_conjunction_double_quoted(*values_duplicate)))

# ....................{ EXCEPTIONS ~ maps                 }....................
@type_check
def die_if_maps_collide(*mappings: MappingType) -> None:
    '''
    Raise an exception if two or more passed dictionaries **collide** (i.e.,
    contain key-value pairs containing the same keys but differing values).

    Parameters
    ----------
    mappings : Tuple[MappingType]
        Tuple of all dictionaries to be validated.

    Raises
    ----------
    BetseMappingException
        If two or more passed dictionaries collide.

    See Also
    ----------
    :func:`is_maps_collide`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import iteriter
    from betse.util.type.text.string import strjoin

    # If two or more of these dictionaries collide...
    if is_maps_collide(*mappings):
        # Iterable of all possible pairs of these dictionaries.
        mappings_pairs = iteriter.iter_pairs(mappings)

        # For each possible pair of these dictionaries...
        for mappings_pair in mappings_pairs:
            # If this pair of dictionaries collides...
            if is_maps_collide(*mappings_pair):
                # Set of all key-value pairs unique to a single mapping.
                items_unique = mappings[0].items() ^ mappings[1].items()

                # Set of all keys visited while iterating this set.
                keys_visited = set()

                # Set of all non-unique keys shared by two or more such pairs.
                keys_collide = set()

                # For each key of such a pair...
                for key, _ in items_unique:
                    # If this key has already been visited, this is a
                    # non-unique key shared by two or more such pairs.
                    if key in keys_visited:
                        keys_collide.add(key)

                    # Mark this key as having been visited.
                    keys_visited.add(key)

                # Grammatically proper noun describing the number of such keys.
                keys_noun = 'key' if len(keys_collide) == 1 else 'keys'

                # Raise an exception embedding this set.
                raise BetseMappingException(
                    'Dictionary {} {} not unique.'.format(
                        keys_noun, strjoin.join_as_conjunction_double_quoted(
                            *keys_collide)))

# ....................{ EXCEPTIONS ~ maps : keys          }....................
@type_check
def die_unless_maps_keys_equal(*mappings: MappingType) -> None:
    '''
    Raise an exception unless all of the passed dictionaries contain the exact
    same keys.

    Equivalently, this function raises an exception if any key of any passed
    dictionary is *not* a key of any other such dictionary.

    Parameters
    ----------
    mappings : Tuple[MappingType]
        Tuple of all dictionaries to be validated.

    Raises
    ----------
    BetseMappingKeyException
        If any key of any passed dictionary is *not* a key of any other such
        dictionary.

    See Also
    ----------
    :func:`is_keys_equal`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    # If one or more of these dictionaries contain differing keys...
    if not is_maps_keys_equal(*mappings):
        # First passed mapping. Since the is_keys_equal() function necessarily
        # returns true if either no mappings or only one mapping are passed,
        # this function returning false implies that two or more mappings are
        # passed. Ergo, this mapping is guaranteed to exist.
        mapping_first = mappings[0]

        # For each mapping excluding the first...
        for mapping in mappings[1:]:
            # If the keys of this mapping differ from those of the first...
            if not is_maps_keys_equal(mapping, mapping_first):
                # Set of all keys differing between these two mappings.
                keys_unequal = mapping.keys().symmetric_difference(
                    mapping_first.keys())

                # Grammatically proper noun describing the number of such keys.
                keys_noun = 'key' if len(keys_unequal) == 1 else 'keys'

                # Raise an exception embedding this set.
                raise BetseMappingKeyException(
                    'Dictionary {} {} differ.'.format(
                        keys_noun, strjoin.join_as_conjunction_double_quoted(
                            *keys_unequal)))

# ....................{ TESTERS ~ map : keys              }....................
@type_check
def has_keys(mapping: MappingType, keys: IterableTypes) -> bool:
    '''
    ``True`` only if the passed dictionary contains *all* passed keys.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be tested.
    keys : IterableTypes
        Iterable of all keys to be tested for.

    Returns
    ----------
    bool
        ``True`` only if this dictionary contains *all* passed keys.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import itertest

    # If any key is unhashable and hence *NOT* a valid key, raise an exception.
    itertest.die_unless_items_instance_of(iterable=keys, cls=HashableType)

    # Return true only if...
    return (
        # If only one key is passed, optimize this common edge case with the
        # standard idiom for testing key containment.
        keys[0] in mapping if len(keys) == 1 else
        # Else, two or more keys are passed. In this case, fallback to a
        # general-case strategy testing key containment in a single one-liner.
        # And yes: this is ridiculously awesome.
        set(keys).issubset(mapping)
    )

# ....................{ TESTERS ~ map : values            }....................
@type_check
def is_values_unique(mapping: MappingType) -> bool:
    '''
    ``True`` only if all values of the passed dictionary are **unique** (i.e.,
    if *no* two key-value pairs of this dictionary share the same values).

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inspected.

    Returns
    ----------
    bool
        ``True`` only if *all* values of this dictionary are unique.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import itertest

    # Defer to an existing low-level tester for sanity.
    return itertest.is_items_unique(mapping.values())

# ....................{ TESTERS ~ maps                    }....................
@type_check
def is_maps_collide(*mappings: MappingType) -> bool:
    '''
    ``True`` only if two or more passed dictionaries **collide** (i.e., contain
    key-value pairs containing the same keys but differing values).

    Equivalently, this function returns ``True`` only if no key of a key-value
    pair of one mapping is also a key of a key-value pair of another mapping
    such that the two values of these two pairs differ.

    Parameters
    ----------
    mappings : Tuple[MappingType]
        Tuple of all dictionaries to be tested.

    Returns
    ----------
    bool
        ``True`` only if if two or more passed dictionaries collide.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable.set import sets

    # If two mappings are passed, prematurely optimize this common case by
    # directly testing these two mappings.
    #
    # Dismantled, this is:
    #
    # * "mappings[0].items() | mappings[1].items()", the set of 2-tuples of all
    #   distinct key-value pairs in these mappings.
    # * "len(mappings[0].items() | mappings[1].items())", the number of all
    #   distinct key-value pairs in these mappings.
    # * "mappings[0].keys()  | mappings[1].keys()", the set of 2-tuples of all
    #   distinct keys in these mappings.
    # * "len(mappings[0].keys()  | mappings[1].keys())", the number of all
    #   distinct keys in these mappings.
    # * "len(...) != len(...)", true only if the number of all distinct
    #   key-value pairs in these mappings differs from the number of distinct
    #   keys in these mappings. If true, then the former is guaranteed to be
    #   strictly larger than the latter (i.e., more distinct key-value pairs
    #   than keys exist), in which case one or more distinct key-value pairs
    #   must necessarily share the same key and hence collide.
    #
    # For example:
    #
    #     # A pair of colliding mappings.
    #     >>> mappings = ({'a': 42, 'b': 24}, {'a': 35, 'c': 53})
    #     >>> items = mappings[0].items() | mappings[1].items()
    #     >>> keys = mappings[0].keys() | mappings[1].keys()
    #     >>> items
    #     {('a', 42), ('a', 35), ('b', 24), ('c', 53)}
    #     >>> keys
    #     {'a', 'b', 'c'}
    #     >>> len(items) > keys
    #     True
    #
    #     # A pair of non-colliding mappings.
    #     >>> mappings = ({'a': 42, 'b': 24}, {'a': 42, 'c': 53})
    #     >>> items = mappings[0].items() | mappings[1].items()
    #     >>> keys = mappings[0].keys() | mappings[1].keys()
    #     >>> items
    #     {('a', 42), ('b', 24), ('c', 53)}
    #     >>> keys
    #     {'a', 'b', 'c'}
    #     >>> len(items) > keys
    #     False
    if len(mappings) == 2:
        return (
            len(mappings[0].items() | mappings[1].items()) !=
            len(mappings[0].keys()  | mappings[1].keys())
        )
    # Else if either no mappings or only one mapping are passed, return true.
    # See the is_maps_keys_equal() function for discussion on this edge case.
    elif len(mappings) < 2:
        return True
    # Else, three or more mappings are passed. In this case, defer to a
    # general-purpose algorithm supporting arbitrarily many mappings.

    # Sets of distinct key-value pairs and keys in these mappings.
    mappings_items = sets.make_union(mapping.items() for mapping in mappings)
    mappings_keys  = sets.make_union(mapping.keys()  for mapping in mappings)

    # Return true only if the number of distinct key-value pairs in these
    # mapping differs from and hence is strictly greater than the number of
    # distinct keys in these mappings. See above for further details.
    return len(mappings_items) != len(mappings_keys)

# ....................{ TESTERS ~ maps                    }....................
@type_check
def is_maps_keys_equal(*mappings: MappingType) -> bool:
    '''
    ``True`` only if all passed dictionaries contain the exact same keys.

    Parameters
    ----------
    mappings : Tuple[MappingType]
        Tuple of all dictionaries to be tested.

    Returns
    ----------
    bool
        ``True`` only if these dictionaries contain the exact same keys.
    '''

    # If two mappings are passed, prematurely optimize this common case by
    # directly testing the keys contained in these two mappings for equality.
    if len(mappings) == 2:
        return mappings[0].keys() == mappings[1].keys()
    # Else if either no mappings or only one mapping are passed, return true.
    # Why? Because:
    #
    # * If no mappings are passed, this edge case is functionally equivalent to
    #   the edge case in which one or more empty mappings (i.e., mappings
    #   containing no key-value pairs) are passed. Since the key containers for
    #   empty mappings are themselves empty, these key containers contain the
    #   exact same keys -- namely, none.
    # * If only one mapping is passed, this mapping by definition contains the
    #   exact same keys as itself.
    elif len(mappings) < 2:
        return True
    # Else, three or more mappings are passed. In this case, defer to a general
    # case algorithm.

    # Keys of the first passed mapping.
    mapping_first_keys = mappings[0].keys()

    # Return true only if...
    return all(
        # This mapping contains the same keys as the first such mapping...
        mapping.keys() == mapping_first_keys
        # For each mapping excluding the first.
        for mapping in mappings[1:]
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all Matplotlib-based animation subclasses.
'''

#FIXME: A potential substantial speedup (albeit possibly non-portable to all
#possible Matplotlib backends) is as follows:
#
#    "One limitation of the methods presented above is that all figure elements
#     are redrawn with every call to draw, but we are only updating a single
#     element. Often what we want to do is draw a background, and animate just
#     one or two elements on top of it. As of matplotlib-0.87, GTKAgg, TkAgg,
#     WXAgg, and FLTKAgg support the methods discussed here.
#     The basic idea is to set the 'animated' property of the Artist you want
#     to animate (all figure elements from Figure to Axes to Line2D to Text
#     derive from the base class Artist). Then, when the standard canvas draw
#     operation is called, all the artists except the animated one will be
#     drawn. You can then use the method background =
#     canvas.copy_from_bbox(bbox) to copy a rectangular region (eg the axes
#     bounding box) into a a pixel buffer. In animation, you restore the
#     background with canvas.restore_region(background), and then call
#     ax.draw_artist(something) to draw your animated artist onto the clean
#     background, and canvas.blit(bbox) to blit the updated axes rectangle to
#     the figure. When I run the example below in the same environment that
#     produced 36 FPS for GTKAgg above, I measure 327 FPS with the techniques
#     below. See the caveats on performance numbers mentioned above. Suffice it
#     to say, quantitatively and qualitiatively it is much faster."
#
#For further details, see:
#
#    https://scipy.github.io/old-wiki/pages/Cookbook/Matplotlib/Animations.html#Animating_selected_plot_elements
#
#As the URL fragment "old-wiki" suggests, there probably exists a newer version
#of this technique ideally generalizable to all possible Matplotlib backends.

#FIXME: All animations should be displayed in a non-blocking rather than
#blocking manner, as required for parallelizing the animation pipeline. To
#minimize memory leaks while doing so, consider responding to animation window
#close events by explicitly closing the current animation on such events: e.g.,
#
#    def _hook_on_close(event) -> None:
#        print('Animation "{}" window closed.'.format(self._kind))
#
#        #FIXME: Is this safe to call here? Presumably, but consider.
#        self._close()
#
#
#    def __init__(...):
#            .
#            -
#            .
#       # Register the close event handler defined above.
#       self._figure.canvas.mpl_connect('close_event', self._hook_on_close)
#            .
#            -
#            .

#FIXME: We should probably animate non-blockingly (e.g., by passing
#"block=False" to the plt.show() command). To do so, however, we'll probably
#have to implement an analogue to Matplotlib's "_pylab_helper.Gcf" global-like
#static class to ensure that object references are retained for all animations
#whose windows have not yet been closed. This implies, in turn, that we'll
#probably have to monkey-patch the appropriate Matplotlib event handlers on
#window close to release these object references. To do so, grep the Matplotlib
#codebase for "Gcf.destroy". Thunderous tales of woe!
#
#Actually, I believe that a simpler approach might be available. Rather
#than implementing yet another "_pylab_helper.Gcf"-like construct, we leverage
#the fact that animation objects should only live as long as their underlying
#figure objects by explicitly adding circular references between the two: e.g.,
#
#    # This is already done by the "VisualCellsABC" superclass.
#    self._figure = pyplot.figure()
#
#    # Then just add this to the AnimCellsABC.__init__() method *BEFORE* the
#    # self._figure.show(block=False) method is called.
#    self._figure.__BETSE_anim__ = self
#
#That said, we might not even need to do that much. Why? Because the
#"FuncAnimation" class is *ALWAYS* passed "self.plot_frame" -- which, being a
#bound method of the current animation object, should ensure that that object
#remains alive. Non-blocking animations may already work out of the box! Oh...
#no. The "FuncAnimation" class is instantiated into an attribute of this class,
#implying that when the last reference to instances of this class go away, they
#everything goes away. We probably will need to add circular references to the
#passed "_figure", as detailed above.
#FIXME: Indeed, a (very minor) amount of research suggests that non-blocking
#animations should be trivially creatable by just ensuring that a reference to
#the instantiated FuncAnimation() object is retained, as intuited above: e.g.,
#
#    https://stackoverflow.com/questions/21099121/python-matplotlib-unable-to-call-funcanimation-from-inside-a-function

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseSimConfException
from betse.lib.matplotlib.matplotlibs import mpl_config
from betse.lib.matplotlib.writer import mplvideo
from betse.lib.matplotlib.writer.mplcls import (
    ImageMovieWriter, NoopMovieWriter)
from betse.science.enum.enumphase import SimPhaseKind
from betse.science.visual.layer.vectorfield.lyrvecfldabc import (
    LayerCellsFieldColorlessABC)
from betse.science.visual.layer.vectorfield.lyrvecfldstream import (
    LayerCellsFieldStream)
from betse.science.visual.visabc import VisualCellsABC
from betse.util.io.log import logs
from betse.util.path import dirs, pathnames
from betse.util.type.iterable import itertest
from betse.util.type.types import type_check, BoolOrNoneTypes, IntOrNoneTypes
from matplotlib import pyplot
from matplotlib.animation import FuncAnimation

# ....................{ BASE                              }....................
class AnimCellsABC(VisualCellsABC):
    '''
    Abstract base class of all animation subclasses.

    Instances of this class animate the spatial distribution of one or more
    modelled variables (e.g., membrane voltage) over all time steps of the
    current simulation.

    Attributes (Private)
    ----------
    _anim : FuncAnimation
        Low-level Matplotlib animation object instantiated by this high-level
        BETSE wrapper object.

    Attributes (Private: Time)
    ----------
    _time_step_count : int
        Number of frames to be plotted.
    _time_step_last : int
        0-based index of the last frame to be plotted, exactly equivalent to
        ``self._time_step_count - 1``.
    _time_step : int
        0-based index of the current frame being plotted, corresponding to the
        0-based sampled time step currently being simulated.
    _time_step_absolute : int
        0-based index of the last frame to be plotted.

    Attributes (Private: Saving)
    ----------
    _save_frame_template : str
        :func:`str.format`-formatted template which, when formatted with the
        0-based index of the current frame, yields the absolute path of the
        image file to be saved for that frame.
    _writer_images : MovieWriter
        Matplotlib object saving animation frames as images if doing so *or*
        ``None`` otherwise.
    _writer_video : MovieWriter
        Matplotlib object saving animation frames as video if doing so *or*
        ``None`` otherwise.

    Attributes (Private: Current)
    ----------
    _is_current_overlayable : BoolOrNoneTypes
        ``True`` only if overlaying either electric current or concentration
        flux streamlines on this animation when requested by the current
        simulation configuration (as governed by the
        ``p.anim.is_overlay_current`` parameter).
    _is_current_overlay_only_gj : bool
        ``True`` only if overlaying intracellular current *or* ``False``
        otherwise (i.e., if overlaying both intra- and extracellular current).
        Ignored unless overlaying current (i.e., if :attr:`_is_current_overlay`
        is ``True``).
    '''

    # ..................{ LIFECYCLE                         }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        save_dir_parent_basename: str,

        #FIXME: Remove the "is_current_overlayable" and
        #"is_current_overlay_only_gj" parameters but *NOT* the corresponding
        #private attributes. Currently, the former parameter is only enabled
        #by subclasses that do *NOT* already plot streamlines. (Makes sense.)
        #Thanks to the new layer API, however, this base class can now
        #automatically detect when the subclass is plotting streamlines by:
        #
        #* Iteratively searching through the "_layers" list in the
        #  animate() function, permitting subclasses to append one or more
        #  layers that plot streamlines.
        #* If any layer in this list is an instance of the new
        #  "LayerCellsFieldStream" subclass, then the "_is_current_overlayable"
        #  attribute should be disabled.
        #* Else, that attribute should default to the corresponding setting(s)
        #  in the current simulation configuration.

        # Optional parameters.
        is_current_overlayable: BoolOrNoneTypes = None,
        is_current_overlay_only_gj: BoolOrNoneTypes = None,
        is_ecm_required: bool = False,
        time_step_count: IntOrNoneTypes = None,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this animation.

        Caveats
        ----------
        If neither displaying nor saving this animation, this animation would
        ideally reduce to a noop. Since this is a superclass method, however,
        simply returning would have little effect. While raising an exception
        would have an effect, doing so would also require all callers to
        explicitly catch and ignore that exception -- in which case this
        animation would hardly have reduced to a noop. In short, ignoring this
        edge case is currently the only sane policy.

        Parameters
        ----------
        save_dir_parent_basename : str
            Basename of the parent directory of the subdirectory to which this
            animation's frames will be saved when requested by the current
            simulation configuration.
        is_current_overlayable : optional[bool]
            ``True`` only if overlaying either electric current or
            concentration flux streamlines on this animation when requested by
            the current simulation configuration (as governed by the
            ``p.anim.is_overlay_current`` parameter). All subclasses except
            those already plotting streamlines (e.g., by calling the superclass
            :meth:`_plot_stream` method) should unconditionally enable this
            boolean.
        is_current_overlay_only_gj : optional[bool]
            ``True`` if only overlaying intracellular current *or* ``False`` if
            overlaying both intra- and extracellular current. Ignored if
            current is _not_ being overlayed at all (i.e., if
            :attr:`_is_current_overlay` is ``False``). If ``None``, defaults to
            the following state:

            * ``False`` if extracellular spaces are enabled _and_ both
               intracellular and extracellular current is being animated.
            * ``True`` if either extracellular spaces are disabled _or_ are
               enabled but only intracellular current is being animated.
        is_ecm_required : optional[bool]
            ``True`` only if this animation is specific to extracellular spaces.
            If ``True`` and extracellular spaces are currently disabled, an
            exception is raised. Defaults to ``False``.
        time_step_count : optional[int]
            Number of frames to be plotted. If `None`, defaults to the number
            of sampled time steps in the current tissue simulation.
        '''

        # Pass all parameters *NOT* listed above to our superclass.
        super().__init__(*args, **kwargs)

        # If this subclass requires extracellular spaces but extracellular
        # spaces are currently disabled, raise an exception.
        if is_ecm_required and not self._phase.p.is_ecm:
            raise BetseSimConfException(
                'Animation "{}" requires extracellular spaces, which are '
                'disabled by the current simulation configuration.'.format(
                self._kind))

        # Default unpassed parameters.
        #
        # If unpassed, overlay streamlines onto this animation only if all of
        # these conditions are satisfied:
        #
        # * The current simulation phase is any phase after the "seed" phase
        #   (i.e., any phase except the "seed" phase). Streamlines require
        #   simulation data defined during the "init" phase and are thus
        #   unavailable in earlier phases.
        # * The current simulation configuration enables this overlay.
        if is_current_overlayable is None:
            is_current_overlayable = (
                self._phase.kind >= SimPhaseKind.INIT and
                self._phase.p.anim.is_overlay_current)
        if is_current_overlay_only_gj is None:
            is_current_overlay_only_gj = not (
                self._phase.p.is_ecm and self._phase.p.IecmPlot)
        if time_step_count is None:
            time_step_count = len(self._phase.sim.time)

        # Classify all remaining parameters.
        self._is_current_overlayable = is_current_overlayable
        self._is_current_overlay_only_gj = is_current_overlay_only_gj
        self._time_step_count = time_step_count

        # 0-based index of the last frame to be plotted.
        self._time_step_last = self._time_step_count - 1

        # Classify attributes to be possibly redefined below.
        self._writer_images = None
        self._writer_video = None

        # 0-based index of the current frame.
        self._time_step = 0

        # If saving animations, prepare to do so.
        self._init_saving(save_dir_parent_basename)


    @type_check
    def _init_saving(self, save_dir_parent_basename: str) -> None:
        '''
        Initialize this animation for platform-compatible file saving if
        enabled by the current simulation configuration _or_ noop otherwise.

        Parameters
        ----------
        save_dir_parent_basename : str
            Basename of the parent directory of the subdirectory to which this
            animation's frames will be saved when requested by the current
            simulation configuration. Defaults to a suitably generic basename.
        '''

        # Animation configuration localized for convenience.
        anim_config = self._phase.p.anim

        # If it is *NOT* the case that...
        if not (
            # This animation is being saved...
            self._is_save and (
            # ...as either images or video.
            anim_config.is_images_save or anim_config.is_video_save)
        # Then this animation is unsaved. In this case, silently noop.
        ):
            return

        #FIXME: This is silly. Rather than prohibiting animation names
        #containing directory separators, simply sanitize this animation's name
        #by globally replacing all such separators with non-separator characters
        #guaranteed to be permitted in pathnames for all platforms (e.g.,
        #replacing "/" with "_" or "-" characters on POSIX systems).

        # If the human-readable name of this animation contains directory
        # separators and hence is *NOT* a valid basename, raise an exception.
        pathnames.die_unless_basename(self._kind)

        # Path of the subdirectory to which these files will be saved,
        # creating this subdirectory and all parents thereof if needed.
        save_dirname = dirs.canonicalize_and_make_unless_dir(pathnames.join(
            self._phase.export_dirname, save_dir_parent_basename, self._kind))

        # If saving animation frames as images, prepare to do so.
        if anim_config.is_images_save:
            #FIXME: This currently defaults to padding frames with six or seven
            #zeroes, on average. Let's make this a bit more aesthetic by
            #padding frames to only as many zeroes are absolutely required by
            #the current frame count. To do that, in turn, we'll probably need
            #to shift everything that follows in this method to the _animate()
            #method, where the actual frame count is finally passed.

            # Template expanding to the basename of each image to be saved.
            # The "ImageMovieWriter" class subsequently expands the "{{"- and
            # "}}"-delimited substring to the 0-based index of the current
            # frame number.
            save_frame_template_basename = '{}_{{:07d}}.{}'.format(
                self._kind, anim_config.image_filetype)

            # Template expanding to the absolute path of each image to be
            # saved.
            writer_images_template = pathnames.join(
                save_dirname, save_frame_template_basename)

            # Object writing animation frames as images.
            self._writer_images = ImageMovieWriter()

            # Log this preparation.
            logs.log_debug(
                'Preparing to save animation frames "%s"...',
                writer_images_template)

            # Prepare to save these animation frames.
            self._writer_images.setup(
                fig=self._figure,
                outfile=writer_images_template,
                dpi=anim_config.image_dpi,
            )

        # If saving animation frames as video, prepare to do so.
        if anim_config.is_video_save:
            # Name of the first video encoder installed on the current system.
            video_writer_name = mplvideo.get_first_writer_name(
                anim_config.video_writer_names)
            # print('found video writer: {}'.format(VideoWriterClass))

            # Matplotlib animation writer class encapsulating this encoder.
            VideoWriterClass = mplvideo.get_writer_class(video_writer_name)

            # Name of the first video codec supported by both this video
            # encoder and the video container format corresponding to this
            # video's filetype.
            video_codec_name = mplvideo.get_first_codec_name(
                writer_name=video_writer_name,
                container_filetype=anim_config.video_filetype,
                codec_names=anim_config.video_codec_names,
            )

            # Basename of the video to be written.
            save_video_basename = '{}.{}'.format(
                self._kind, anim_config.video_filetype)

            # Absolute path of the video to be written.
            writer_video_filename = pathnames.join(
                save_dirname, save_video_basename)

            # Object writing animation frames as video. Matplotlib animation
            # writer classes must *NOT* be manually instantiated, as doing so
            # fails to apply the monkey-patching applied by this function.
            self._writer_video = mplvideo.make_writer(
                cls=VideoWriterClass,
                bitrate=anim_config.video_bitrate,
                codec=video_codec_name,
                fps=anim_config.video_framerate,
                metadata=anim_config.video_metadata,
            )

            # Log this preparation.
            logs.log_debug(
                'Preparing to save animation video "%s"...',
                writer_video_filename)

            # Prepare to save this animation video. Matplotlib squelches
            # critical (technically non-fatal but effectively fatal)
            # warnings and errors emitted by the external command invoked
            # by this call to the MovieWriter.setup() method *UNLESS* the
            # current matplotlib-specific verbosity level is "debug".
            # Temporarily ensure this for the duration of this call.
            with mpl_config.reducing_log_level_to_debug_if_info():
                self._writer_video.setup(
                    fig=self._figure,
                    outfile=writer_video_filename,
                    dpi=anim_config.video_dpi,
                )

    # ..................{ PROPERTIES                        }..................
    # Read-only properties, preventing callers from resetting these attributes.

    @property
    def time_step(self) -> int:
        '''
        0-based index of the current frame being plotted, corresponding to the
        0-based sampled time step currently being simulated.
        '''

        return self._time_step

    # ..................{ PREPARERS                         }..................
    # This method has been overridden to support subclasses that manually
    # handle animations rather than calling the _animate() method (e.g., the
    # "AnimCellsWhileSolving" subclass).
    def _prep_figure(self, *args, **kwargs) -> None:

        #FIXME: Refactor this method as follows:
        #
        #* Eliminate all current overlay-specific boolean attributes (e.g.,
        #  "_is_current_overlayable"), as detailed in an __init__() method
        #  comment above.
        #* Shift the _prep_layer_current() method into the superclass.
        #* Call the _prep_layer_current() method in the superclass
        #  _prep_figure() implementation.
        #* Remove this _prep_figure() implementation.

        # Append a layer overlaying current density *AFTER* all lower layers
        # (e.g., cell data) have been appended but *BEFORE* all higher layers
        # (e.g., cell labelling) have been appended.
        self._prep_layer_current()

        # Perform superclass figure preparation.
        super()._prep_figure(*args, **kwargs)


    def _prep_layer_current(self) -> None:
        '''
        Append a layer overlaying current density if this simulation
        configuration requests a current overlay and no layer in the layer
        sequence already plots streamlines _or_ noop otherwise.

        If an existing layer in the layer sequence already plots streamlines,
        this method avoids adding another layer also doing so. Since the
        current overlay also plots streamlines, attempting to plot streamlines
        over existing streamlines would produce an unintelligible plot or
        animation. (That would be bad.)
        '''

        # If either...
        if (
            # This simulation configuration requests no current overlay.
            not self._is_current_overlayable or

            # A layer in the layer sequence already plots a vector field.
            itertest.is_item_instance_of(
                iterable=self._layers, cls=LayerCellsFieldColorlessABC)
        # ...then silently noop.
        ):
            return
        # Else, overlay current.

        # Current density vector field to layer.
        field = None

        # If layering intracellular current, do so.
        if self._is_current_overlay_only_gj:
            logs.log_debug('Overlayering intracellular current...')
            field = self._phase.cache.vector_field.currents_intra
        # Else, layer extracellular current.
        else:
            logs.log_debug('Overlayering extracellular current...')
            field = self._phase.cache.vector_field.currents_extra

        # Append a layer overlaying this field.
        self._append_layer(LayerCellsFieldStream(field=field))

    # ..................{ ANIMATORS                         }..................
    @type_check
    def _animate(self, *args, **kwargs) -> None:
        '''
        Display and/or save this animation as requested by the current
        simulation configuration.

        This method is intended to be called as the last statement in the
        :meth:`__init__` method of all subclasses of this superclass.

        Parameters
        ----------
        All parameters are passed to the :meth:`_prep_figure` method.
        '''

        # Prepare for plotting immediately *BEFORE* plotting the first frame.
        self._prep_figure(*args, **kwargs)

        #FIXME: For efficiency, we should probably be passing "blit=True," to
        #FuncAnimation(). Unfortunately, doing so will necessitate
        #restructuring animations to conform to blitting-specific requirements,
        #including:
        #
        #* The definition of a new init_frame() method of this class, which
        #  plots all portions of this animation *NOT* changing between frames.
        #  Are there any? We have no idea. Axes ticks presumably never change
        #  between frames for any animation, so there should be *SOMETHING* to
        #  plot here.
        #* The passing of this method to the "FuncAnimation" instance defined
        #  below via the "init_func" parameter: e.g.,
        #      init_func=self.init_frame,
        #* Refactoring the plot_frame() method to return the tuple of all
        #  matplotlib artist objects modified for the current frame. This is
        #  probably infeasible in a generic manner given the current crusty
        #  design of this class, but should be trivial (...in theory) after
        #  redesigning this class to support composoble
        #
        #    http://devosoft.org/making-efficient-animations-in-matplotlib-with-blitting
        #
        #Lemon grass and dill!

        # Create and assign an animation function to a local variable. If the
        # latter is *NOT* done, this function will be garbage collected prior
        # to subsequent plot handling -- in which case only the first plot will
        # be plotted without explicit warning or error. Die, matplotlib! Die!!!
        self._anim = FuncAnimation(
            # Figure to which the "func" callable plots each frame.
            fig=self._figure,

            # Callable plotting each frame.
            func=self.plot_frame,

            # Number of frames to be animated.
            frames=self._time_step_count,

            #FIXME: The interval should, ideally, be synchronized with the FPS
            #used for video encoding. To guarantee this:
            #
            #* Generalize the FPS option in the configuration file to *ALL*
            #  animations. Currently, this option only applies to video
            #  encoding.
            #* Convert the currently configured FPS into this interval in
            #  milliseconds as follows:
            #
            #      interval = 1000.0 / fps

            # Delay in milliseconds between consecutive frames. To convert this
            # delay into the equivalent frames per second (FPS):
            #
            #      fps = 1000.0 / interval
            interval=200,

            #FIXME: This is a bit silly. Ideally, animations should *ALWAYS* be
            #repeatable. Once we've refactored away usage of the
            #Animation.save() method, refactor:
            #
            #* This parameter to unconditionally enable repeating: e.g.,
            #      repeat=True,
            #* The plot_frame() method to conditionally call MovieWriter
            #  methods (e.g., grab_frame(), finish()) *ONLY* if the current
            #  call to the plot_frame() method is the first such call for the
            #  current frame. While this state would be trivial for this class
            #  to record, perhaps matplotlib's "Animation" base class already
            #  records this state? Contemplate us up.

            # Indefinitely repeat this animation unless saving animations, as
            # doing so under the current implementation would repeatedly (and
            # hence unnecessarily) overwrite previously written files.
            repeat=not self._is_save,
        )

        try:
            # If displaying and optionally saving this animations, do so.
            if self._is_show:
                #FIXME: If the current backend is non-interactive (e.g.,
                #"Agg"), the following function call reduces to a noop. This is
                #insane, frankly. In this case, this animation's plot_frame()
                #is never called! No errors or warnings are logged, so it's
                #unclear who or what is the culprit here. If the pyplot.show()
                #function is indeed only supported by interactive backends, we
                #should do the following here:
                #
                #* Detect whether or not the current backend is
                #  non-interactive.
                #* If so, either:
                #  * Emit an explicit warning advising the user that this
                #    animation will almost certainly be silently ignored. This
                #    isn't terribly ideal, but it's better than zilch.
                #  * If this animation is currently being saved, simply perform
                #    the non-display save logic performed below -- which *DOES*
                #    behave as expected for non-interactive backends. Clearly,
                #    the culprit is the pyplot.show() function. Mournful sigh.

                # Display and optionally save this animations. Note that,
                # although this function is called in a blocking manner, the
                # GUI-driven event loops of some interactive backends appear to
                # ignore the request for blocking behavior and perform
                # non-blocking behaviour instead. This, in turn, prevents this
                # branch from reliably finalizing this animation by calling the
                # close() method. This differs from the non-interactive
                # saving-specific branch that follows, which is guaranteed to
                # behave in a blocking manner and hence *CAN* reliably call the
                # close() method. tl;dr: GUIs, so random.
                pyplot.show()
            # Else if only saving but not displaying this animation *AND* at
            # least one animation writer doing so is enabled, do so.
            elif self._is_save and (
                self._writer_images is not None or
                self._writer_video is not None
            ):
                # Save this animation by iteratively calling our plot_frame()
                # method to save each animation frame. Since this method
                # already manually saves each such frame for the case of both
                # displaying *AND* saving this animation via the above call to
                # the pyplot.show() function, that logic is reused here by
                # preventing this call to the Animation.save() method from
                # attempting to automatically save each such frame.
                #
                # By default, the Animation.save() method iteratively calls the
                # MovieWriter.grab_frame() method of the passed "writer" object
                # to save each such frame. If no such object is passed, this
                # object defaults to a new writer whose name is the current
                # value of the "animation.writer" rcparam. Hence, there exists
                # no means of preventing the Animation.save() method from
                # writing. However, there also exists no good alternative to
                # this method that iteratively calls our plot_frame() method
                # without also writing. For example:
                #
                # * The pyplot.show() function iterating frames silently
                #   reduces to a noop for non-interactive backends and is thus
                #   inapplicable as a general-purpose solution.
                # * The frame iteration automatically performed by the
                #   Animation.save() method is both non-trivial and requires
                #   calls to private methods of the matplotlib animation API.
                #   While this iteration could (and actually was, in the first
                #   implementation of this approach) be reduplicated here,
                #   doing so would be overly fragile and hence break under
                #   upstream changes to this private API.
                #
                # The robust solution is to instead pass the Animation.save()
                # method a writer reducing to a noop, circumventing conflicts
                # with the manual saving performed by our plot_frame() method.
                self._anim.save(
                    # Note that, since "NoopMovieWriter" maintains no state, a
                    # singleton "NoopMovieWriter" instance could technically be
                    # shared amongst all animation classes. However, since
                    # "NoopMovieWriter" construction is trivially fast, there
                    # are no demonstrable advantages and arguable disadvantages
                    # to doing so (e.g., code complexity, space consumption).
                    writer=NoopMovieWriter(),

                    # Pass an ignorable filename. To guarantee that an
                    # exception is raised on this method attempting to read or
                    # write this file, pass a filename guaranteed to be invalid
                    # on all supported platforms (e.g., containing null bytes).
                    # For understandable reasons, this parameter is mandatory.
                    filename=pathnames.INVALID_PATHNAME,
                )

                # Finalize saving this animation.
                self.close()

        # plt.show() unreliably raises exceptions on window close resembling:
        #     AttributeError: 'NoneType' object has no attribute 'tk'
        # This error appears to ignorable and hence is caught and squelched.
        except AttributeError as exc:
            # If this is that exception, mercilessly squelch it.
            if str(exc) == "'NoneType' object has no attribute 'tk'":
                pass
            # Else, reraise this exception.
            else:
                raise

    # ..................{ CLOSERS                           }..................
    def close(self) -> None:
        '''
        Finalize displaying and/or saving this non-blocking animation.

        This method is intended to be called as the last statement in the
        `_animate()` method animating this non-blocking subclass. If this
        subclass is blocking, this method must _not_ be called.

        See Also
        ----------
        :meth:`VisualCellsABC.close`
            Superclass method destroying this animation's low-level plot and
            all memory associated with this plot.
        '''

        # Finalize this animation's low-level plot.
        super().close()

        # Finalize all writers saving this animation if any.
        self._close_writers()

        # Prevent this animation from being reused *AND* break hard cycles.
        self._anim = None


    def _close_writers(self) -> None:
        '''
        Finalize all writers saving this animation if any.
        '''

        # If saving animation frames as images...
        if self._writer_images is not None:
            # Finalize doing so.
            self._writer_images.finish()

            # Prevent this writer from being reused and break hard cycles.
            self._writer_images = None

        # If saving animation frames as video...
        if self._writer_video is not None:
            # Finalize doing so. For debuggability, temporarily escalate the
            # matplotlib-specific verbosity level.
            with mpl_config.reducing_log_level_to_debug_if_info():
                self._writer_video.finish()

            # Prevent this writer from being reused and break hard cycles.
            self._writer_video = None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for matplotlib, a mandatory runtime dependency.

Backends
--------
matplotlib supports numerous **interactive backends** (i.e., bindings to
external GUI-specific widget toolkits), only one of which will be imported by
matplotlib at runtime. If the caller specifies no such backend, a default
backend specific to the current system will be imported. However, all such
backends including such defaults are fairly fragile and hence prone to raising
exceptions under common contexts.

The following table summarizes our current findings:

=========  ========  ======  ========  ======  ========  ======  ========
Backend    Footnote  Is Supported Under?
                     ----------------------------------------------------
                     Linux             OS X              Windows
                     ----------------  ----------------  ----------------
                     Frozen  Unfrozen  Frozen  Unfrozen  Frozen  Unfrozen
=========  ========  ======  ========  ======  ========  ======  ========
CocoaAgg   1         No      No        No      No        No      No
Gtk3Agg    2         No      No        No      No        No      No
Gtk3Cairo  3         No      No        ???     ???       ???     ???
MacOSX               No      No        Yes     Yes       No      No
Qt4Agg     3         No      No        ???     ???       ???     ???
Qt5Agg               ???     ???       ???     ???       ???     ???
TkAgg                Yes     Yes       No      Yes       ???     ???
=========  ========  ======  ========  ======  ========  ======  ========

Footnote descriptions are as follows:

1. Backend "CocoaAgg" is officially deprecated and known to be broken.
2. Backend "Gtk3Agg" is known to be broken under Python 3.
3. These backends do *not* support our current animation method, despite
   otherwise working (e.g., to display static plots).
'''

#FIXME: Refactor backend_names() to discover backend names via the standard
#module "pkg_utils" rather than by manually delving through the filesystem,
#which fails under frozen executables.

#FIXME: It'd be great to raise human-readable exceptions on the specified
#backends *NOT* being available. This is certainly feasible, as the
#following stackoverflow answer demonstrates -- if somewhat involved:
#    https://stackoverflow.com/questions/5091993/list-of-all-available-matplotlib-backends
#That said, we really want to do this *ANYWAY* to print this list when running
#"betse info". So, let's just get this done, please.

#FIXME: Consider contributing most or all of this submodule back to matplotlib.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To allow matplotlib defaults (e.g., backend, logging) to be replaced
# with application-specific preferences, no matplotlib module may be imported
# until *AFTER* the MplConfig.init() method has been called -- including here
# at the top-level.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys
from beartype.typing import Generator  # ContextManager
from betse.exceptions import BetseMatplotlibException
from betse.util.io.error import errexception
from betse.util.io.log import logs
from betse.util.io.log.conf import logconf
from betse.util.io.log.logenum import LogLevel
from betse.util.os import (
    displays,
    kernels,
)
from betse.util.os.brand import macos
from betse.util.py.module import pymodname
from betse.util.type.iterable import itersort
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict
from betse.util.type.numeric import versions
from betse.util.type.types import (
    type_check, MappingType, SequenceTypes, SetType, StrOrNoneTypes,)
from contextlib import contextmanager
from logging import Logger

# ....................{ GLOBALS ~ str                      }....................
_BACKEND_NAME_HEADLESS = 'Agg'
'''
Name of the non-GUI-based matplotlib backend to fallback to in the event that
*NO* GUI-based matplotlib backend is usable on this system. For portability,
this backend is guaranteed to be usable on all platforms and systems regardless
of matplotlib version.
'''

# ....................{ GLOBALS ~ dict                     }....................
_LOG_LEVEL_TO_VERBOSITY_LEVEL_NAME = {
    LogLevel.NONE:     'silent',
    LogLevel.CRITICAL: 'silent',
    LogLevel.ERROR:    'silent',
    LogLevel.WARNING:  'silent',
    LogLevel.INFO:     'helpful',
    LogLevel.DEBUG:    'debug',
    LogLevel.ALL:      'debug-annoying',
}
'''
Dictionary mapping from each standard logging level expected by newer
matplotlib versions (i.e., matplotlib >= 2.2.0) to the corresponding
non-standard verbosity level name expected by older matplotlib versions (i.e.,
matplotlib < 2.2.0).
'''


_RC_PARAMS = {
    # This option's default value of 20 is slightly too small for our use case,
    # inviting unctuous non-fatal runtime warnings resembling:
    #
    #    /builds/betse/betse/conda-env/lib/python3.7/site-packages/matplotlib/pyplot.py:514:
    #    RuntimeWarning: More than 20 figures have been opened. Figures created
    #    through the pyplot interface (`matplotlib.pyplot.figure`) are retained
    #    until explicitly closed and may consume too much memory. (To control
    #    this warning, see the rcParam `figure.max_open_warning`).
    #
    # While these warnings may be entirely eliminated by setting this option to
    # 0, doing so has the unfortunate side effect of failing to notify both
    # developers and end users of imminent memory exhaustion. To err on the
    # side of caution, we thus instead reduce the frequency of this warning.
    'figure.max_open_warning': 50,
}
'''
Dictionary mapping ``matplotlibrc`` option names to corresponding values.

The :meth:`MplConfig.init` method subsequently updates
the :attr:`matplotlib.rcParams` dictionary of default ``matplotlibrc`` options
deserialized from the current ``matplotlibrc`` file with this dictionary. Ergo,
the custom options specified by this dictionary override the default options
defined by that file.
'''

# ....................{ CLASSES                            }....................
class MplConfig(object):
    '''
    High-level wrapper simplifying low-level configuration and introspection of
    matplotlib.
    '''

    # ..................{ INITIALIZERS                       }..................
    def init(self, backend_name: StrOrNoneTypes) -> None:
        '''
        Reconfigure matplotlib with sane defaults specific to the current
        system.

        On first importation, matplotlib configures itself by loading the
        contents of the first ``matplotlibrc`` file found in any of several
        candidate directories. Technically, this application *could* supply an
        application-specific version of this file to force matplotlib to adopt
        application-specific configuration settings. Since synchronizing this
        local copy with remote changes is an onerous (if not ultimately
        infeasible) chore, we elect instead to reconfigure matplotlib *after*
        this file has already been loaded at application startup. While this
        does increase startup costs, the alternatives are all absurd at best.

        Parameters
        ----------
        backend_name: optional[str]
            Name of the matplotlib backend to explicitly enable. Defaults to
            ``None``, in which case this method implicitly enables the first
            importable backend known to be both usable and supported by
            application requirements (in descending order of preference).

        See Also
        ----------
        http://matplotlib.org/users/customizing.html
            Matplotlib configuration details.
        '''

        # Avoid circular import dependencies.
        from betse.lib.matplotlib import mplcolormap

        # Initialize matplotlib in a safe manner.
        self._init_matplotlib()

        # Establish the matplotlib backend *AFTER* initializing matplotlib.
        self._init_backend(backend_name=backend_name)

        # Register all custom colormaps *AFTER* initializing matplotlib.
        mplcolormap.init()

        # Import all animation writer classes *AFTER* establishing the
        # matplotlib backend, thus implicitly registering these classes with
        # matplotlib. This permits other code in the codebase to conveniently
        # refer to these classes by name without having to manually instantiate
        # these classes. (While it may or may not be necessary to import these
        # classes after establishing the backend, it is only prudent to do so.)
        from betse.lib.matplotlib.writer import mplcls
        if False: mplcls    # silence contemptible IDE warning messages


    def _init_matplotlib(self) -> None:
        '''
        Initialize matplotlib in a safe manner, guaranteeing all subsequent
        importations of matplotlib elsewhere in the codebase to be safe.

        This method imports matplotlib in a safe manner and should be the first
        call to do so for the active Python process. Unfortunately, the
        :mod:`matplotlib.__init__` submodule implicitly imported on the first
        matplotlib importation performs the following unsafe logic:

        * The current command-line argument list :data:`sys.argv` is
          iteratively searched for arguments matching patterns, including:
          * Arguments prefixed by ``-d`` of length greater than or equal to 3
            (e.g., ``-dtkagg`` but _not_ simply `-d`). For each such argument,
            :mod:`matplotlib.__init__` attempts to enable the backend whose
            name is given by such argument excluding the prefixing ``-d``,
            silently ignoring exceptions.
          * Arguments prefixed by ``--verbose-`` matching a matplotlib-specific
            verbosity level (e.g., ``--verbose-debug``). For each such
            argument, :mod:`matplotlib.__init__` coerces the global verbosity
            to that level.

        This is utterly horrible. Since enabling arbitrary backends can have
        non-negligible side effects, the :mod:`matplotlib.__init__` submodule
        *must* be prevented from performing this logic. Since this submodule is
        imported only on the first importation of a matplotlib module, it
        performing this preventation *only* on the first importation of the
        top-level :mod:`matplotlib` package by this method suffices to sanitize
        matplotlib behaviour.
        '''

        # Copy the current argument list into a temporary list.
        _sys_argv_old = sys.argv[:]

        # Remove all arguments following the mandatory basename of the current
        # process from the current argument list, preventing matplotlib from
        # inspecting arguments and hence enabling arbitrary backends.
        del(sys.argv[1:])

        # Import the "matplotlib.__init__" module. Since this module prints
        # verbose messages if and only if the matplotlib-specific CLI option
        # "--verbose-debug" is in this argument list, this option must be
        # explicitly added to this list *BEFORE* importing this module. Awful!
        try:
            # Logging configuration singleton.
            log_config = logconf.get_log_conf()

            # Logging level of the application-wide standard output handler,
            # representing the default logging level for this Python process.
            betse_log_level = log_config.handler_stdout.level

            # Logging level of matplotlib itself, applicable *ONLY* to
            # matplotlib >= 2.2.0. See comments below for contemptible details.
            matplotlib_log_level = None

            # Name of the matplotlib-specific verbosity level corresponding to
            # this numeric level to the corresponding, applicable *ONLY* to
            # matplotlib < 2.2.0.
            #
            # Since there exist fewer matplotlib-specific verbosity levels than
            # standard logging lovels, this conversion is necessarily lossy. A
            # hearty thanks for failing to conform to well-established Python
            # standards yet again, matplotlib!
            #
            # Note that matplotlib-specific verbosity levels have been
            # obsoleted as of matplotlib 2.2.0 by the standard "logging" API.
            # Ergo, all of the logic below pertaining to this variable is
            # required *ONLY* to ensure backward compatibility with older
            # matplotlib versions (i.e., matplotlib < 2.2.0).
            #
            # Ideally, this logic would be conditionally performed *ONLY* if
            # the current version of matplotlib is sufficiently old.
            # Unfortunately, attempting to do so raises chicken-and-egg issues;
            # notably, the current version of matplotlib can only be determined
            # by importing the top-level "matplotlib" module, but doing so
            # requires *ALL* matplotlib-specific CLI options (including those
            # setting verbosity) to have already been injected into the
            # "sys.argv" list.  Since these two constraints are mutually
            # exclusive *AND* since squelching verbosity in older versions of
            # matplotlib takes precedence over doing so more intelligently, we
            # do so unintelligently (i.e., unconditionally) -- regardless of
            # whether this version of matplotlib even supports these options.
            matplotlib_log_level_name = None

            # If BETSE-specific debugging is enabled, enable the mildest form
            # of matplotlib-specific logging. While matplotlib also supports
            # the "debug" and "debug-annoying" verbosity levels, both produce
            # far more effluvia than is generally useful.
            if betse_log_level <= LogLevel.DEBUG:
                matplotlib_log_level = LogLevel.INFO
                matplotlib_log_level_name = 'helpful'
            # Else, squelch all matplotlib-specific logging. Since matplotlib
            # supports no fine-grained verbosity levels between "helpful" and
            # "silent", reduce all non-debugging levels to merely "silent".
            else:
                matplotlib_log_level = LogLevel.WARNING
                matplotlib_log_level_name = 'silent'

            # Convert this name into a matplotlib-specific CLI option.
            # print('matplotlib verbosity: ' + matplotlib_log_level_name)
            # matplotlib_log_level_name='debug'
            sys.argv.append('--verbose-' + matplotlib_log_level_name)

            #FIXME: We should additionally set the "ffmpeg"-specific CLI option
            #"-loglevel" based on the above "betse_log_level" as well --
            #perhaps by setting "rcParams['animation.ffmpeg_args'] = '-report'"
            #or some such *AFTER* importing matplotlib below.

            # Log this initialization.
            logs.log_debug(
                'Initializing matplotlib with options: %s', sys.argv)

            # Set matplotlib's logging level *BEFORE* importing from
            # matplotlib. Since matplotlib now sanely defaults to the WARNING
            # level, this could technically also be performed *AFTER* importing
            # from matplotlib. Doing so would ignore all messages logged during
            # this importation with a lower level. (That would be bad.)
            #
            # Note that this logic effectively reduces to a noop under older
            # matplotlib versions (i.e., matplotlib < 2.2.0). Although
            # non-ideal, this appears to have *NO* harmful side effects.
            #
            # Ideally, the following two lines would reduce to this one-liner:
            #
            #     self.log_level = matplotlib_log_level
            #
            # Unfortunately, the property setter implicitly invoked by such an
            # assignment assumes matplotlib to already have been initialized --
            # which, of course, it hasn't. (Chicken-and-egg issues, people.)
            self.logger.setLevel(matplotlib_log_level)

            # Import matplotlib submodules *AFTER* setting all matplotlib CLI
            # options, which this importation internally parses in older
            # versions of matplotlib as a non-idempotent side effect.
            import matplotlib      # this is horrible
            from matplotlib import rcParams  # , font_manager

            # If the "matplotlib.verbose" object exists in this version,
            # monkeypatch this object to prevent the
            # matplotlib.verbose.set_level() method from reducing to a noop, as
            # occurs when this private attribute is *NOT* nullified. waat?
            #
            # For reasons, matplotlib 2.2.0 critically broke backward
            # compatibility in an apocalyptic manner by removing this object.
            # For discussion, see the following matplotlib issue:
            #
            #     https://github.com/matplotlib/matplotlib/issues/10716
            if hasattr(matplotlib, 'verbose'):  # this is horrible, too!
                matplotlib.verbose._commandLineVerbose = None
        # Guarantee the prior argument list to be restored from this temporary
        # list even in the event of exceptions.
        finally:
            sys.argv = _sys_argv_old
            del(_sys_argv_old)

        # Unconditionally enable all settings defined by this global.
        #
        # Note that calling the rcParams.update() method appears to be the only
        # reliable means of overriding default matplotlib options. See also a
        # relevant StackOverflow response:
        #     https://stackoverflow.com/a/34377868/2809027
        rcParams.update(_RC_PARAMS)

        #FIXME: Sadly, the "font_manager.USE_FONTCONFIG" global is currently
        #only modifiable by physically modifying the contents of the
        #"matplotlib/font_manager.py" file at installation time. Yes, this is
        #patently insane -- but such are the vagaries of a pale life.

        # If the external "fc-match" command is in the current ${PATH}, the
        # POSIX-compliant fontconfig subsystem for portably locating system
        # fonts is available. Since fontconfig is the canonical means of
        # performing font lookups under most POSIX-compliant platforms (e.g.,
        # Linux, macOS), instruct matplotlib to do so. For unknown reasons,
        # matplotlib's fontconfig support is labelled "experimental" and hence
        # disabled by default. Since this support appears to be sufficiently
        # robust for our limited use case, manually enable this support.
        # if pathables.is_pathable('fc-match'):
        #     font_manager.USE_FONTCONFIG = True


    #FIXME: Revise docstring in accordance with the iterative platform-specific
    #fallback process now employed for deciding the default Matplotlib backend.
    def _init_backend(self, backend_name: StrOrNoneTypes) -> None:
        '''
        Set the default matplotlib backend to be implicitly used for subsequent
        plotting, which matplotlib requires to be configured *before* the first
        importation of any the following modules: :mod:`matplotlib.pyplot` or
        :mod:`matplotlib.backends`.

        Specifically:

        * If this method is called by :mod:`pytest`-based testing and hence
          possibly by headless continuous integration (CI) with no access to a
          window manager, the non-interactive `Agg` backend is used.
        * Else, if the current platform is:
          * Either Linux or Windows, the ``TkAgg`` backend is used. This is the
            only backend currently known to survive freezing into executables.
            Alternatives include:
            * ``Qt4Agg``, an (arguably) aesthetically inferior backend with
              (inarguably) significant performance concerns.
          * OS X, the ``MacOSX`` backend is used. This is the only backend
            currently known to survive freezing into executables. Alternatives
            include:
            * ``CocoaAgg``, a non-native backend leveraging the cross-platform
              C++ library AGG (Anti-grain Geometry). Yes! This backend is
              officially deprecated and fundametally broken, however. No!

        Parameters
        ----------
        backend_name: optional[str]
            Name of the matplotlib backend to explicitly enable. Defaults to
            ``None``, in which case this method implicitly enables the first
            importable backend known to be both usable and supported by
            application requirements (in descending order of preference).
        '''

        # Avoid circular import dependencies.
        from betse.util.type.text.string import strjoin

        # If no specific backend is requested *AND* the active Python process
        # is headless and hence supports only headless backends...
        if backend_name is None and displays.is_headless():
            # Log this observation.
            logs.log_info(
                'Headless display environment detected. '
                'Defaulting to headless backend "%s"...',
                _BACKEND_NAME_HEADLESS)

            # Default to a headless backend.
            backend_name = _BACKEND_NAME_HEADLESS

        # If a specific backend is requested...
        if backend_name is not None:
            # If this backend is usable, enable this backend and return. Note
            # that, as the is_backend_usable() method internally enables this
            # backend as a necessary side effect of testing this backend's
            # usability, this backend need not (and, indeed, should not) be
            # explicitly re-enabled here.
            if self.is_backend_usable(backend_name):
                return
            # Else, log a non-fatal error and continue.
            else:
                logs.log_error(
                    'Preferred backend "%s" unusable.', backend_name)
        # Else, no specific backend is requested. Since the default headless
        # backend (e.g., "Agg") should *ALWAYS* be usable, this typically
        # implies this process to be headfull and hence support GUI backends.
        # Default to the first backend usable under the current system.

        # Name of the current platform (e.g., "Linux", "Darwin", "Windows").
        kernel_name = kernels.get_name()

        # Tuple of the names of all matplotlib backends to iteratively
        # fallback to on this platform if supported or None otherwise.
        backend_names = self._kernel_name_to_backend_names_prefer.get(
            kernel_name, None)

        # If this platform is unsupported, raise an exception.
        if backend_names is None:
            raise BetseMatplotlibException(
                'Platform "{}" unsupported.'.format(kernel_name))

        # Log this iteration.
        logs.log_debug('Finding usable backend in: %r', backend_names)

        #FIXME: If this is Matplotlib >= 2.0.0 and the only available backend
        #is "TkAgg", a non-fatal warning should be logged instructing the user
        #to install a more stable backend (e.g., "Qt5Agg", "Qt4Agg", "WxAgg").
        #To leverage the local "backend_names" list, this warning should
        #probably be emitted here rather than in the is_backend_usable()
        #method.

        # For each such backend (in descending order of preference)...
        for backend_name in backend_names:
            # If this backend is usable, this tester has already implicitly
            # enabled this backend. Our work is done here.
            if self.is_backend_usable(backend_name):
                return
        # Else, no preferred GUI-based backend is usable on the current system
        # (e.g., due to no external widget library being installed).

        # If the fallback non-GUI-based backend is usable...
        if self.is_backend_usable(_BACKEND_NAME_HEADLESS):
            # Log a non-fatal warning.
            logs.log_warning(
                'No usable GUI-based matplotlib backend found. '
                'Defaulting to usable CLI-based backend "%s". '
                'Consider installing support for GUI-based backends %s.',
                _BACKEND_NAME_HEADLESS,
                strjoin.join_as_disconjunction_double_quoted(*backend_names))

            # Default to this backend and return.
            self.backend_name = _BACKEND_NAME_HEADLESS
            return

        # Else, no backends appear to be usable on the current system. Due to
        # the ubiquity of the headless fallback backend (e.g., "Agg"), this
        # should only occur in extremely uncommon edge cases (e.g., manual
        # recompilation of the entire system). If this does occur, this is
        # typically the result of internal issues in this codebase. To assist
        # in debugging these issues, a simple exception message is raised.
        raise BetseMatplotlibException(
            'No usable matplotlib backend found. '
            '{}-supported backends tested include (in order): {}.'.format(
                kernel_name,
                strjoin.join_as_conjunction_double_quoted(*backend_names)))

    # ..................{ PROPERTIES                         }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # NOTE: To avoid desynchronization issues between low-level matplotlib
    # internals and the following high-level properties wrapping these
    # internals, most of these properties leverage the less efficient but safer
    # @property decorator rather than the more efficient but less safe
    # @property_cached decorator.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    @property_cached
    def version(self) -> str:
        '''
        Currently installed version of matplotlib as a ``.``-delimited version
        specifier (e.g., ``2.1.0``).
        '''

        # Delay importation of the "matplotlib.__init__" module.
        import matplotlib
        from betse.util.type.text import regexes

        # Currently installed version of matplotlib to be returned.
        version_return = matplotlib.__version__

        # Either:
        #
        # * If this version exhibits formatting discrepancies introduced in
        #   recent matplotlib versions violating Python conventions for version
        #   specifiers (e.g., the non-compliant version "2.1.0-python3_4"
        #   rather than the compliant version ""2.1.0"), a sequence whose
        #   single item is this version stripped of these discrepencies.
        # * Else, "None".
        #
        # Note that this regular expression need *NOT* be compiled for
        # efficiency, as this property is cached and hence called only once.
        version_munged_or_none = regexes.get_match_groups_numbered_or_none(
            text=version_return, regex=r'^([^-]+)-')

        # Else, this version is non-compliant. In this case, reduce this to a
        # compliant version.
        if version_munged_or_none is not None:
            logs.log_debug(
                'Sanitizing matplotlib version from "%s" to "%s"...',
                version_return, version_munged_or_none[0])
            version_return = version_munged_or_none[0]

        # Return this possibly munged version.
        return version_return

    # ..................{ PROPERTIES ~ log                   }..................
    @property
    def logger(self) -> Logger:
        '''
        Matplotlib-specific logger.

        Caveats
        -----------
        This property effectively reduces to a noop under older matplotlib
        versions (i.e., matplotlib < 2.2.0), which ignored the stardarized
        logging system in favour of a matplotlib-specific verbosity toggle.
        Although non-ideal, this appears to have *no* harmful side effects.
        '''

        # One-liners for the glory of flying spaghetti monsters.
        return logs.get_logger('matplotlib')

    # ..................{ PROPERTIES ~ log : level           }..................
    @property
    def log_level(self) -> LogLevel:
        '''
        Matplotlib-specific logging level.

        If this is a sufficiently old version of matplotlib *not* supporting
        the :mod:`logging` API (i.e., matplotlib < 2.2.0), this method attempts
        to internally convert from a matplotlib-specific verbosity level name
        to a :mod:`logging` level. If this name is:

        * ``silent``, hiding all Matplotlib output *except* non-fatal warnings,
          :attr:`LogLevel.WARNING` is returned. Not exactly silent, is it?
        * ``helpful``, emitting terse Matplotlib debugging output,
          :attr:`LogLevel.INFO` is returned.
        * `debug`, emitting verbose Matplotlib debugging output,
          :attr:`LogLevel.DEBUG` is returned.
        * `debug-annoying`, emitting *very* verbose Matplotlib debugging
          output. :attr:`LogLevel.ALL` is returned. You *never* want this.
          (Trust us.)
        '''

        # Return this logger's level.
        #
        # Since the matplotlib-specific logger unconditionally sets this
        # logger's level regardless of whether the current matplotlib version
        # actually logs to this logger, this logging level is returned as is.
        # Hence, the above docstring is an elaborate lie. (Ignore it, please.)
        return self.logger.level


    @log_level.setter
    @type_check
    def log_level(self, log_level: LogLevel) -> None:
        '''
        Set the current matplotlib-specific logging level to the passed level.

        Parameters
        -----------
        log_level: LogLevel
            Matplotlib-specific logging level to be set.
        '''

        # Set the matplotlib-specific logger's level to the passed level.
        self.logger.setLevel(log_level)

        # Delay importation of the "matplotlib.__init__" module. To ensure
        # messages logged by this importation are logged with the passed level,
        # we do so *AFTER* setting this level.
        import matplotlib

        # If the "matplotlib.verbose" object exists in this version, internally
        # convert the passed level to an obsolete verbosity level name required
        # by this obsolete version of matplotlib. *SIGH*
        matplotlib_verbose = getattr(matplotlib, 'verbose', None)
        if matplotlib_verbose is not None:  # this is horrible
            # Non-standard verbosity level name corresponding to the passed
            # level expected by older matplotlib versions (i.e., matplotlib <
            # 2.2.0) *OR* "None" if no such name corresponds.
            verbosity_level_name = _LOG_LEVEL_TO_VERBOSITY_LEVEL_NAME.get(
                log_level, None)

            # If this name is unrecognized, raise an exception. For unknown
            # reasons, the Verbose.set_level() method called below unsafely
            # emits non-fatal warnings rather than raising fatal exceptions on
            # receiving an unrecognized level name.
            if (
                verbosity_level_name is None or
                verbosity_level_name not in matplotlib_verbose.levels
            ):
                raise BetseMatplotlibException(
                    'Matplotlib verbosity level "{}" unrecognized.'.format(
                        verbosity_level_name))

            # Set this verbosity level.
            matplotlib_verbose.set_level(verbosity_level_name)

    # ..................{ PROPERTIES ~ path                  }..................
    @property
    def cache_dirname(self) -> str:
        '''
        Absolute pathname of the platform- and typically user-specific
        directory to which matplotlib caches metadata (e.g., about fonts).
        '''

        # Delay importation of the "matplotlib.__init__" module.
        import matplotlib

        # Return this path.
        return matplotlib.get_cachedir()


    @property
    def rc_filename(self) -> str:
        '''
        Absolute pathname of the currently selected ``matplotlibrc`` file
        establishing default matplotlib options.
        '''

        # Delay importation of the "matplotlib.__init__" module.
        import matplotlib

        # Return this path.
        return matplotlib.matplotlib_fname()

    # ..................{ PROPERTIES ~ backend               }..................
    @property
    def backend(self) -> type(sys):
        '''
        In-memory submodule object corresponding to the current backend.

        Raises
        ----------
        BetseMatplotlibException
            If no backend has been set yet.
        '''

        # If no backend has been set yet, raise an exception.
        if not self.is_backend():
            raise BetseMatplotlibException('Matplotlib backend not yet set.')

        # Fully-qualified name of this backend's submodule.
        backend_module_name = (
            'matplotlib.backends.backend_' + self.backend_name)

        # This backend's submodule. Since this backend has been set, this
        # submodule *SHOULD* still be cached in-memory. Let's be sure.
        backend_module = sys.modules.get(backend_module_name, None)
        if backend_module is None:
            raise BetseMatplotlibException(
                'Current matplotlib backend module "{}" not imported.'.format(
                    backend_module_name))

        # Return this submodule.
        return backend_module


    @property
    def backend_canvas_class(self) -> type(object):
        '''
        :class:`matplotlib.backend_bases.FigureCanvasBase` subclass unique to
        the current backend (e.g.,
        :class:`matplotlib.backends.backend_qt5.FigureCanvasQT` for the
        ``qt5agg`` backend).

        Note that, whereas each backend implements a backend-specific canvas,
        each backend shares the same :class:`matplotlib.figure.Figure` class
        (i.e., there exist *no* backend-specific figure subclasses).

        Raises
        ----------
        BetseMatplotlibException
            If no backend has been set yet.
        '''

        # This backend's module.
        backend = self.backend

        # This subclass. While all standard backends *SHOULD* set the module
        # attribute "FigureCanvas" to the "FigureCanvas" subclass specific to
        # that backend, let's be sure.
        backend_canvas_class = getattr(backend, 'FigureCanvas', None)
        if backend_canvas_class is None:
            raise BetseMatplotlibException(
                'Matplotlib backend canvas class '
                '"{}.FigureCanvas" not found.'.format(backend.__name__))
        return backend_canvas_class


    @property
    def backend_figure_filetypes(self) -> list:
        '''
        List of all figure filetypes supported by the current backend (e.g.,
        `['bmp', 'gif', 'jpg', 'png', 'tiff']`).

        If no backend has been set yet, an exception is raised.
        '''

        # There exist two means of acquiring this metadata:
        #
        # 1. Creating an empty canvas specific to the current backend.
        # 2. Creating an empty figure specific to the current backend.
        #
        # Since the latter also implicitly requires creating an empty canvas
        # and is hence both more expensive and more complex, we prefer the
        # former.
        backend_canvas_class = self.backend_canvas_class

        # The "FigureCanvasBase" superclass of all canvas subclasses defines
        # the public class attribute "filetypes" to be the dictionary of all
        # filetypes supported by that canvas. Moreover, all subclasses reliably
        # redefine this attribute in the expected way. While the same list is
        # also obtainable by creating an instance of this subclass and calling
        # that subclass' get_supported_filetypes() method, the current approach
        # is substantially more efficient.
        #
        # Magic is magic. Do not question magic, for it is magical.
        return list(backend_canvas_class.filetypes.keys())

    # ..................{ PROPERTIES ~ backend : name        }..................
    @property
    def backend_name(self) -> str:
        '''
        Human-readable name (e.g., ``Qt5Agg``) of the current backend.

        This name is *not* guaranteed to be lowercase. This name is typically a
        mix of lower- and uppercase alphanumeric characters.

        This backend is *not* guaranteed to be usable. If no backend has been
        enabled yet (either explicitly or implicitly), this is the name of the
        default backend implicitly enabled by importing the
        :mod:`motplotlib.pyplot` submodule.
        '''

        # Delay importation of the "matplotlib.__init__" module.
        import matplotlib

        # Return this backend's name.
        return matplotlib.get_backend()


    @backend_name.setter
    @type_check
    def backend_name(self, backend_name: str) -> None:
        '''
        Set the current backend to the usable backend with the passed name.

        This name is interpreted case-insensitively and hence may be in any
        case including lower- and uppercase (e.g., `tkagg`, `TKAGG`, `TkAgg`).

        This backend is assumed to be usable (i.e., the
        :meth:`is_backend_usable` method is assumed to succeed when passed this
        name). If this is *not* the case, matplotlib will internally raise an
        exception of indeterminate origin and readability. Welcome to hell.
        '''

        # Log this attempt.
        logs.log_debug('Enabling matplotlib backend "%s"...', backend_name)

        # If enabling the "TkAgg" backend *AND* this is matplotlib >= 2.0.0,
        # log a prominent warning. This backend is known to behave unstably
        # under all newer versions of matplotlib. See the
        # "_kernel_name_to_backend_names_prefer" poperty for further details.
        if (
            backend_name == 'TkAgg' and
            versions.is_greater_than_or_equal_to(self.version, '2.0.0')
        ):
            logs.log_warning(
                'Matplotlib backend "TkAgg" known to be unstable '
                'under matplotlib >= 2.0.0.')

        # Attempt to enable this backend.
        try:
            self._enable_backend(backend_name)
        # If a presumably non-human-readable exception was raised...
        except Exception:
            # Log a human-readable error first.
            logs.log_error(
                'Matplotlib backend "%s" not found or unusable.', backend_name)

            # Re-raise this exception.
            raise

    # ..................{ PROPERTIES ~ backend : names       }..................
    @property_cached
    def backend_names(self) -> SequenceTypes:
        '''
        Sequence of the strictly lowercase names of all currently available
        matplotlib-specific backends (e.g.,
        ``['gtk3agg', 'tkagg', 'qt4agg']``).

        This property effectively yields the contents of the canonical
        :attr:`matplotlib.rcsetup.all_backends` sequence such that all items
        are both lowercased and lexicographically sorted.
        '''

        # Importing this submodule may have side effects and is thus deferred.
        from matplotlib.backends import backend_registry

        # List of all publicly supported backends, defined as the concatenation
        # of these subsidiary lists:
        # * "interactive_bk", listing all supported interative backends.
        # * "non_interactive_bk", listing all supported non-interative
        #   backends.
        #
        # Note that this list excludes numerous privately supported backends
        # residing in the "matplotlib.backends" subpackage (e.g.,
        # "matplotlib.backends.backend_webagg_core"). Obsolete versions of
        # Matplotlib permitted callers to enable these backends; ergo, this
        # property was previously implemented to dynamically introspect the
        # local filesystem to collect the set of all backend names from the set
        # of all "backend_"-prefixed submodules residing in this subpackage.
        # Modern versions of Matplotlib prohibit callers from enabling these
        # backends; ergo, this property now defers to this sequence.
        backend_names = backend_registry.list_builtin()

        # Return a new sequence containing all items from "all_backends"
        # lowercased and lexicographically sorted.
        return itersort.sort_ascending(
            backend_name.lower() for backend_name in backend_names)

    # ..................{ PROPERTIES ~ backend : names: pri  }..................
    @property_cached
    def _backend_names_blacklist(self) -> SetType:
        '''
        Unordered set of the names of all **blacklisted backends** (i.e.,
        backends known to be dangerously unusable on the current system).

        The :meth:`is_backend_usable` method directly reports these backends to
        be unusable *without* unsafely testing this to be the case. Attempting
        to programmatically test the unusability of these backends is known to
        induce extreme fatal errors in the worst case, including segnmentation
        faults with no human-readable exception messages.

        These backends include:

        * ``Gtk3Agg``, emitting the following non-fatal warning when enabled:

            UserWarning: The Gtk3Agg backend is not known to work on Python
            3.x.

        * ``Gtk3cairo``, which, despite claiming to be a GTK+ 3.x-specific
          backend, appears to attempt to dynamically load GTK+ 2.x-specific
          shared libraries -- inducing the same segmentation fault as above.
        * All GTK+ 2.x-specific backends (e.g., ``Gtk``, ``Gtkagg``),
          conflicting with GTK+ 3.x-specific backends (e.g., ``Gtk3``,
          ``Gtk3agg``). Attempting to switch to the latter after having already
          switched to the former typically induces the following fatal
          segmentation fault immediately halting the current process:

            (betse:5000): Gtk-ERROR **: GTK+ 2.x symbols detected. Using GTK+
            2.x and GTK+ 3 in the same process is not supported
            zsh: trace trap (core dumped)  betse info

          This is *not* a high-level Python exception and hence cannot be
          caught from within Python. This is a POSIX-level process signal.

        * All Qt 4-specific backends (e.g., ``Qt4Agg``, ``Qt4cairo``). Qt 4 has
          been officially unsupported by the Qt Company (QTC) since 2015 -- and
          hence increasingly insecure, error-prone, and unavailable.
        * If the current platform is *not* macOS:

          * ``macosx``, a Carbon-specific backend usable only under macOS.

        * If the active Python interpreter is running **headless** (i.e., with
          *no* access to a GUI display, often due to running remotely over an
          SSH-encrypted connection supporting only CLI input and output):

          * All Qt 5-specific backends (e.g., ``Qt5Agg``, ``Qt5cairo``). When
            headless, attempting to switch to *any* Qt backend (including both
            Qt 4- and 5-specific backends) induces a silent fatal segmentation
            fault immediately halting the current process.

        Sometimes, the only winning move is not to play at all.
        '''

        # Set of all backends to be unconditionally blacklisted.
        backend_names_blacklist = {
            # Blacklist all GTK+ 2.x-specific backends.
            'gtk', 'gtkagg', 'gtkcairo',

            # Blacklist all GTK+ 2.x-specific backends.
            'gtk3', 'gtk3agg', 'gtk3cairo',

            # Blacklist all Qt 4-specific backends.
            'qt4agg', 'qt4cairo',
        }

        # If the current platform is *NOT* macOS, blacklist the "macosx"
        # backend usable only under macOS.
        if not macos.is_macos():
            backend_names_blacklist.add('macosx')

        # If headless, blacklist the "Qt5Agg" backend known to silently
        # terminate the current process under obscure edge cases.
        if displays.is_headless():
            backend_names_blacklist.update(('qt5agg', 'qt5cairo'))

        # Return this set.
        return backend_names_blacklist


    @property_cached
    def _kernel_name_to_backend_names_prefer(self) -> MappingType:
        '''
        Dictionary mapping from the name of each supported platform to a
        sequence of the names of all matplotlib backends that the
        :meth:`_init_backend` method iteratively defers to on that platform (in
        descending order of preference) in the event the end user fails to
        explicitly set this name (e.g., via the `--matplotlib-backend` option).

        In that case, that method defers to the first matplotlib backend usable
        on the current system whose name is in this tuple.

        This property is efficiently created and cached on the first access.
        '''

        # List of the names of all backends known to be unconditionally
        # supported by this application across *ALL* supported platforms.
        #
        # These are (in descending order of preference):
        #
        # * "Qt5Agg", a GUI backend with (arguably) superior aesthetics but
        #   (inarguably) significant performance concerns by compare to more
        #   preferable backends. Also, note that enabling the experimental
        #   non-blocking behaviour via "pyplot.show(block=False)" reliably
        #   induces segmentation faults with *NO* exception traceback under
        #   at least Windows. Something is better than nothing, though.
        # * "Qt4Agg", a GUI backend with (arguably) inferior aesthetics and
        #   (inarguably) significant performance concerns by compare to more
        #   preferable backends.
        # * "WxAgg", a GUI backend with (arguably) inferior aesthetics and
        #   (inarguably) significant performance *AND* stability concerns by
        #   compare to more preferable backends. As of this writing, the
        #   much-vaunted Python 3.x version of WxPython (i.e., Phoenix) has
        #   only recently been released and supported by matplotlib. Issues
        #   are likely to plague this backend for decades.
        backend_names_prefer = ['Qt5Agg', 'Qt4Agg', 'WxAgg',]

        # If this is matplotlib >= 2.0.0, deprioritize the "TkAgg" backend.
        #
        # For unknown reasons, all recent versions of matplotlib have
        # fundamentally broken this backend with respect to non-blocking
        # animations -- either with or without experimental non-blocking
        # behaviour. This has been extensively tested and isolated to
        # matplotlib itself, which... is frustrating. Let me tell you,
        # matplotlib: we are displeased and we know where your repo lives.
        if versions.is_greater_than_or_equal_to(self.version, '2.0.0'):
            # Append this backend to the end of this list.
            backend_names_prefer.append('TkAgg')
        # Else, prioritize the "TkAgg" backend. Why? Because this backend has
        # adequate (albeit not particularly impressive) aesthetics and superior
        # performance compared to all other backends under older versions of
        # matplotlib. Tcl/Tk: who would've thought?
        else:
            # Prepend this backend to the beginning of this list.
            backend_names_prefer.insert(0, 'TkAgg')

        # Darwin-specific list of such names, prioritizing the only truly
        # usable Darwin-specific backend before the platform-agnostic backends.
        # Since Darwin and Linux are both POSIX-compatible, cross-platform
        # backends (e.g., "Qt5Agg") tend to behave similarly on both platforms.
        backend_names_prefer_darwin = ['MacOSX',] + backend_names_prefer

        # Under Windows, prefer the exact same backends as preferred under
        # Linux. While POSIX-incompatible and hence irregular, Windows still
        # supports the same backends preferred under Linux in the same order.
        backend_names_prefer_windows = backend_names_prefer

        # Dictionary to be returned and cached.
        kernel_name_to_backend_names_prefer = {
            'Linux':   backend_names_prefer,
            'Darwin':  backend_names_prefer_darwin,
            'Windows': backend_names_prefer_windows,
        }

        # Return and cache this dictionary
        return kernel_name_to_backend_names_prefer

    # ..................{ TESTERS ~ backend                  }..................
    def is_backend(self) -> bool:
        '''
        ``True`` only if a backend has been successfully enabled.

        This function returns ``True`` only if a backend has been either:

        * Explicitly enabled by setting the :meth:`backend_name` property
          internally calling the :func:`matplotlib.use` function.
        * Implicitly enabled by importing either of the following submodules:
          * :mod:`matplotlib.pyplot`.
          * :mod:`matplotlib.backends`.
        '''

        # While ridiculous, this test corresponds exactly to the test performed
        # by the matplotlib.use() method itself to detect repetitious calls.
        # Since the "matplotlib.pyplot" subpackage internally imports this
        # subpackage, testing only this subpackage suffices.
        return pymodname.is_imported('matplotlib.backends')


    @type_check
    def is_backend_usable(self, backend_name: str) -> bool:
        '''
        ``True`` only if the backend with the passed name is usable on the
        current system.

        Specifically, this method returns ``True`` only if this backend:

        * May be switched to *without* raising exceptions.
        * May both create and destroy a hidden (i.e., invisible) empty figure
          *without* raising exceptions.

        If this backend is usable, this method switches the current backend to
        this backend *without* restoring the previously set backend. This is
        the unavoidable price of robust, reproducible test results. Callers
        requiring the previously set backend to be restored must do so manually
        (e.g., by setting the :func:`property` attribute to the name of that
        backend) *after* calling this method.
        '''

        # Log this attempt.
        logs.log_debug(
            'Testing matplotlib backend "%s" usability...', backend_name)

        # Lowercase this name, ensuring case-insensitive backend names.
        backend_name = backend_name.lower()

        #FIXME: Overkill and uninformative. To blacklist backends properly, what
        #we *REALLY* want to do instead is to:
        #
        #* Remove the "_backend_names_blacklist" property.
        #* Remove this conditional block entirely.
        #* Improve the _enable_backend() method to initially:
        #  * Define a local dictionary "BLACKLIST_BACKEND_NAME_TO_REASON"
        #    mapping from the names of all blacklisted backend names to
        #    human-readable exception messages describing exactly why those
        #    backends have been blacklisted.
        #  * Test whether the passed backend name is a key of this dictionary.
        #  * If so, raise an exception embedding the corresponding value of this
        #    dictionary.
        #
        #Yeah. That's blatantly heaps better. Hindsight is always 20-20, neh?

        # If this backend is unconditionally blacklisted, report this backend
        # to be unusable *WITHOUT* unsafely attempting to switch to this
        # backend.
        if backend_name in self._backend_names_blacklist:
            logs.log_debug(
                'Matplotlib backend "%s" unusable, '
                'due to being internally blacklisted.', backend_name)
            return False

        # Attempt to...
        try:
            # Enable this backend.
            self._enable_backend(backend_name)

            # If doing so succeeds, this backend is usable.
            logs.log_debug('Matplotlib backend "%s" usable.', backend_name)
            return True
        # If doing so raises an exception, this backend is unusable.
        except Exception as exception:
            # Traceback for this exception.
            traceback = errexception.get_traceback(exception)

            # Log this traceback for debuggability.
            logs.log_debug(
                'Matplotlib backend "%s" unusable, '
                'due to raising the following exception:\n%s',
                backend_name, traceback)

            # Report this backend to be unusable.
            return False

    # ..................{ TESTERS ~ block                    }..................
    def is_backend_current_nonblockable(self) -> bool:
        '''
        ``True`` only if the current backend supports true non-blocking display
        on the current system.

        See Also
        ----------
        :meth:`is_backend_nonblockable`
            Further details.
        '''

        return self.is_backend_nonblockable(self.backend_name)


    @type_check
    def is_backend_nonblockable(self, backend_name: str) -> bool:
        '''
        ``True`` only if the backend with the passed name supports **true
        non-blocking display** (i.e., in a backend-specific GUI eventloop in
        a separate thread of the current process) on the current system.

        If this method returns:

        * ``True``, this backend may reliably display plots and animations in a
          non-blocking manner by passing the experimental ``block=False``
          parameter to the :func:`matplotlib.pyplot.show` method.
        * ``False``, the experimental ``block=False`` parameter must *not* be
          passed to the :func:`matplotlib.pyplot.show` function. Doing so will
          either silently hide plots and animations with no warnings or errors
          *or* raise a low-level segmentation fault with no high-level
          exception traceback. Instead, consider falling back to the following
          approach:

          .. code:: python

              # Setup the current plot or animation for non-blocking display.
              matplotlib.interactive(True)
              matplotlib.pyplot.show()

              # Display the current frame of this plot or animation.
              matplotlib.pyplot.draw()
              matplotlib.pyplot.pause(0.0001)

              # Teardown the current plot or animation.
              matplotlib.interactive(False)

        Caveats
        ----------
        The fallback approach detailed above typically emits the following
        non-fatal warning on the first ``matplotlib.interactive(True)`` call:

        .. code::

           [betse] /usr/lib64/python3.4/site-packages/matplotlib/backend_bases.py:2437:
           MatplotlibDeprecationWarning: Using default event loop until
           function specific to this GUI is implemented
             warnings.warn(str, mplDeprecation)

        While verbose, this warning appears to be safely ignorable.
        '''

        # For safety, a whitelist approach is preferred. That is, only backends
        # explicitly known to support experimental "pyplot.show(block=False)"
        # behavior are assumed to do so; all other backends are assumed to
        # *NOT* support this behavior.
        return backend_name == 'TkAgg'

    # ..................{ GETTERS                            }..................
    def get_rc_param(self, param_name) -> object:
        '''
        Value of the parameter with the passed ``.``-delimited name (e.g.,
        ``savefile.dpi``) in the external ``matplotlibrc`` file.

        If no such parameter exists, an exception is raised.

        Raises
        ----------
        KeyError
            If no such parameter exists.
        '''

        # Delay importation of the "matplotlib.__init__" module.
        from matplotlib import rcParams

        # Return this parameter's value.
        return rcParams[param_name]


    def get_metadata(self) -> OrderedArgsDict:
        '''
        Ordered dictionary synopsizing the current matplotlib installation.
        '''

        # This dictionary.
        metadata = OrderedArgsDict(
            'rc file', self.rc_filename,
            'cache dir', self.cache_dirname,
            'current backend', self.backend_name,
        )

        # For each available backend, add metadata synopsizing that backend.
        for backend_name in self.backend_names:
            metadata[
                'backend {} usable'.format(backend_name.capitalize())] = str(
                    self.is_backend_usable(backend_name)).lower()

        # Get this dictionary.
        return metadata

    # ..................{ CONTEXTS                           }..................
    @contextmanager
    def reducing_log_level_to_debug_if_info(self) -> Generator:
        '''
        Context manager setting the matplotlib-specific verbosity level to
        :attr:`LogLevel.DEBUG` if currently :attr:`LogLevel.INFO` for the
        duration of this context.

        This context manager temporarily increases this level by one level.
        Although the :attr:`LogLevel.INFO` and :attr:`LogLevel.DEBUG` levels
        *both* produce debug output, only the latter produces debug output for
        external commands invoked by matplotlib (e.g., for encoding video via
        ``ffmpeg``); the former produces *no* such output. Since the latter is
        overly verbose for general use and hence useful only for specific
        cases, consider instead:

        * Defaulting to the :attr:`LogLevel.INFO` level.
        * Escalating to the :attr:`LogLevel.DEBUG` level by explicitly entering
          this context manager for the duration of special-case work requiring
          verbosity.

        This context manager guaranteeably reverts this level to the prior
        level even when fatal exceptions are raised. If this level is *not*
        currently :attr:`LogLevel.INFO`, this context manager is a noop.

        Returns
        -----------
        contextlib._GeneratorContextManager
            Context manager setting the level as described above.

        Yields
        -----------
        None
            Since this context manager yields no value, the caller's ``with``
            statement must be suffixed by *no* ``as`` clause.
        '''

        # If the current level is "helpful"...
        if self.log_level is LogLevel.INFO:
            # Escalate to the "debug" level temporarily.
            self.log_level = LogLevel.DEBUG

            # Yield control to the body of the caller's "with" block.
            try:
                yield
            # Revert to the prior level even if that block raised an exception.
            finally:
                self.log_level = LogLevel.INFO
        # Else, the current level is *NOT* "helpful". Reduce to a noop.
        else:
            yield

    # ..................{ MAKERS                             }..................
    #FIXME: Define a new make_backend_figure_manager() method as well. This
    #method is fine, for now. We might very well want to call this sometime.
    def make_backend_figure(self, *args, **kwargs):
        '''
        Create and return a new :class:`Figure` instance associated with a new
        :class:`FigureCanvasBase` instance corresponding to the current backend
        (e.g., :class:`FigureCanvasQt5Agg` for the `qt5agg` backend), passed
        the passed arguments.

        This figure will _not_ be associated with a new or existing
        :class:`FigureManagerBase` instance corresponding to the current
        backend (e.g., :class:`FigureManagerQt5Agg` for the `qt5agg` backend)
        _or_ with the :mod:`matplotlib.pyplot` interface. In theory, this
        implies that this figure should be implicitly garbage collected on:

        * If either _not_ displayed or displayed in a blocking manner, leaving
          scope.
        * If displayed in a non-blocking manner, this figure's window being
          closed.

        If no backend has been set yet, an exception is raised.

        Returns
        ----------
        matplotlib.figure.Figure
            Such figure.
        '''

        # Delay importation of the "matplotlib.__init__" module.
        from matplotlib.figure import Figure

        # Create this figure.
        figure = Figure(*args, **kwargs)
        self.backend_figure_canvas(figure)

        # Return this figure.
        return figure

    # ..................{ PRIVATE                            }..................
    @type_check
    def _enable_backend(self, backend_name: str) -> None:
        '''
        Set the current backend to the backend with the passed name.

        This name is interpreted case-insensitively and hence may be in any
        case including lower- and uppercase (e.g., `tkagg`, `TKAGG`, `TkAgg`).

        This low-level method is principally intended to be called by
        higher-level methods (e.g., :meth:`backend_name`,
        :meth:`is_backend_usable`) wrapping this method with additional
        caller-friendly logic.
        '''

        # Delay importation of the "matplotlib.__init__" module.
        import matplotlib

        # Lowercase this name, as backend names are case-insensitive.
        backend_name = backend_name.lower()

        # True only if a backend has already been enabled.
        is_backend = self.is_backend()

        # Wrap all attempts to import from unsafe matplotlib subpackages (e.g.,
        # "matplotlib.backends", "matplotlib.pyplot") with logic safely undoing
        # these imports when *NO* backend has already been set. (See below.)
        try:
            # If no backend has been enabled...
            if not is_backend:
                # Log this attempt.
                logs.log_debug(
                    'Enabling matplotlib backend "%s" via use()...',
                    backend_name)

                # Enable this backend by calling the use() function, which is
                # expected to behave in a stable manner.
                matplotlib.use(backend_name)
            # Else, a backend has already been enabled.
            #
            # If this backend is tbe currently enabled backend, noop. This is
            # substantially safer than calling the switch_backend() function in
            # the subsequent conditional branch, which is known to be unstable.
            elif backend_name == self.backend_name:
                # Log this attempt.
                logs.log_debug(
                    'Ignoring already enabled matplotlib backend "%s".',
                    backend_name)

                # Reduce to a noop.
                return
            # In this unfortunate case, we have no recourse but to call the
            # switch_backend() function known to be unstable.
            else:
                # Log this attempt.
                logs.log_debug(
                    'Enabling matplotlib backend "%s" via switch_backend()...',
                    backend_name)

                # Unfortunately, if the current platform is macOS *AND* the new
                # backend to be enabled is "TkAgg", enabling this backend is
                # *NOT* safe and must absolutely be prohibited by raising an
                # exception.  Attempting to enable this backend under this edge
                # case commonly results in a segmentation fault, terminating
                # the active Python process in a non-human-readable manner
                # resembling:
                #
                #     [betse] Testing matplotlib backend "tkagg"...
                #     backend TkAgg version 8.5
                #     2016-12-31 01:53:08.886 Python[19521:163945] -[NSApplication _setup:]: unrecognized selector sent to instance 0x7fc278fbec60
                #     2016-12-31 01:53:08.890 Python[19521:163945] An uncaught exception was raised
                #     2016-12-31 01:53:08.890 Python[19521:163945] -[NSApplication _setup:]: unrecognized selector sent to instance 0x7fc278fbec60
                #     2016-12-31 01:53:08.891 Python[19521:163945] (
                #              0   CoreFoundation                      0x00007fff936ea452 __exceptionPreprocess + 178
                #              ...
                #              97  ???                                 0x0000000000000004 0x0 + 4
                #     )
                #     2016-12-31 01:53:08.892 Python[19521:163945] *** Terminating app due to uncaught exception 'NSInvalidArgumentException', reason: '-[NSApplication _setup:]: unrecognized selector sent to instance 0x7fc278fbec60'
                #     *** First throw call stack:
                #     (
                #             0   CoreFoundation 0x00007fff936ea452 __exceptionPreprocess + 178
                #             ...
                #             97  ???  0x0000000000000004 0x0 + 4
                #     )
                #     libc++abi.dylib: terminating with uncaught exception of type NSException
                #     Abort trap: 6
                #
                # This is a common issue afflicting numerous matplotlib users under
                # all versions of macOS. The core issue appears to be that the
                # Xcode-bundled installation of Tcl/Tk is *NOT* usable by Python.
                # There exist two solutions (in no particular order):
                #
                # 1. Reinstall Python to use a Homebrew- or MacPorts-compiled
                #    installation of Tcl/Tk instead. While this does constitute
                #    a valid solution for end users, BETSE itself has no means
                #    of enforcing this dictate and *MUST* thus assume the
                #    current installation of Tcl/Tk to be the Xcode-bundled
                #    version.
                # 2. Call "matplotlib.pyplot.use('TkAgg')" *BEFORE* the first
                #    import of the "matplotlib.pyplot" submodule. While BETSE
                #    itself could technically attempt to enforce this by
                #    preferentially detecting the "TkAgg" backend *BEFORE* all
                #    other backends, the "MacOS" backend is always preferable
                #    under macOS and should thus always be detected first.
                #
                # In short, no sane solution exists. The only sane solution is
                # to refuse to play the game at all.
                if backend_name == 'tkagg' and macos.is_macos():
                    raise BetseMatplotlibException(
                        'Matplotlib backend "TkAgg" not '
                        'safely switchable to under macOS.')

                # Delay importation of this submodule until *ABSOLUTELY*
                # necessary.  Importing this submodule implicitly enables the
                # default matplotlib backend defined by the "backend" RC
                # parameter in the current matplotlib configuration if no
                # backend has been enabled, which can exhibit harmful side
                # effects in common edge cases.
                from matplotlib import pyplot

                # Switch from the current to the passed backend.
                pyplot.switch_backend(backend_name)

            # In either case, this backend has now been set. Since this does
            # *NOT* imply this backend to be usable, further testing is needed.

            # Delay importation of this submodule until *ABSOLUTELY* necessary.
            # See pertinent commentary above.
            from matplotlib import pyplot

            # Validate the usability of this backend by attempting to create
            # and destroy a hidden empty figure. Technically, doing so could
            # exhibit harmful side effects on uncooperative platforms (e.g.,
            # Windows) in common edge cases.
            #
            # Unfortunately, doing so is also essential. The success of
            # switching to this backend above is a necessary but insufficient
            # condition of this backend's usability. While the success of
            # switching to some backends (e.g., "TkAgg") does reliably imply
            # those backends to be usable, the success of switching to other
            # backends (e.g., "Qt5Agg") only implies that the corresponding
            # packages (e.g., "PyQt5") are successfully importable; it does
            # *NOT* imply these packages and thus these backends to actually be
            # usable in any sense. Notably, the PyQt family of backends are
            # infamous for raising non-human-readable exceptions on attempting
            # to display the first figure. "Why, PyQt? WHY!?!?"
            pyplot.figure()
            pyplot.close()
        # If an exception is raised, this backend is unusable,
        except:
            # If no backend was set before the current call to this method,
            # unimport all top-level matplotlib subpackages internally
            # imported by the above call. Doing so ensures that the next
            # call to the is_backend() method reports False and hence that
            # this same conditional branch is re-entered on the next call to
            # this method.
            #
            # This is *NOT* merely a caller convenience. This is critical
            # functionality required for sane testing of backend usability.
            # In particular, if this is *NOT* done, the next call to this
            # method is guaranteed to raise a non-human-readable exception.
            # Why? Because, in that case:
            #
            # * The modules unimported here will still have been imported.
            # * Importing this module sets the current matplotlib backend,
            #   either to the backend requested by the first call to the
            #   matplotlib.use() function if called or to the default
            #   matplotlib backend set by "matplotlibrc" otherwise. In
            #   either case, a backend has now been set.
            # * Since that call to the matplotlib.use() function raised an
            #   exception, that backend is unusable. If that backend is the
            #   default matplotlib backend (e.g., "Qt5Agg"), then the
            #   default matplotlib backend is also unusable.
            # * The "matplotlib.pyplot" subpackage uses the current
            #   matplotlib backend if the "matplotlib.backends" subpackage
            #   has been imported.
            # * Since the latter has indeed been imported, the subsequent
            #   attempt to import the former will re-raise the same
            #   exception as initially raised here.
            # * Ergo, this method may only be safely called once.
            #
            # Since the above behaviour is insane, this subpackage is
            # unimported instead to sanitize life.
            if not is_backend:
                pymodname.unimport_module_if_imported(
                    'matplotlib.backends', 'matplotlib.pyplot')

            # Re-raise this exception.
            raise

# ....................{ SINGLETONS                         }....................
mpl_config = MplConfig()
'''
Singleton matplotlib configuration wrapper.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level module and package importation facilities.

This submodule *only* defines functions implementing non-standard and
occasionally risky "black magic" fundamentally modifying Python's standard
importation semantics and mechanics. This submodule does *not* define
commonplace functions for dynamically importing modules or testing or
validating that importation.

See Also
----------
:mod:`betse.util.py.module.pymodname`
    Related submodule defining functions importing modules by name as well as
    testing and validating that importation.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory dependencies,
# the top-level of this module may import *ONLY* from packages guaranteed to
# exist at installation time -- which typically means *ONLY* BETSE packages and
# stock Python packages.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys
from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ CONVERTERS                        }....................
def to_str_modules_imported_name() -> str:
    '''
    Human-readable string of the fully-qualified names of all previously
    imported modules for the active Python interpreter.

    For readability, these names are sorted in lexicographic order and
    delimited by a newline.
    '''
    # Avoid circular import dependencies.
    from betse.util.type.iterable.mapping import mappings

    # Return the human-readable string produced by joining on newline a
    # generator comprehension yielding the colon-delimited name and value of
    # all environment variables sorted in lexicographic order.
    #
    # Note that the "environ" object is of non-standard type "os._Environ",
    # which the pprint.pformat() fails to recognize and hence format as a
    # "dict"-compatible mapping. Ergo, passing "environ" directly to the
    # mappings.to_str_flat() function would yield a non-human-readable string.
    # While this could be ameliorated by converting "environ" to a "dict" first
    # (e.g., "mappings.to_str_flat(dict(environ))"), doing so still produces
    # less human-readable output than the current approach.
    return mappings.to_str_flat(sys.modules)

# ....................{ REGISTRARS                        }....................
@type_check
def register_dir(dirname: str) -> None:
    '''
    Register all files and subdirectories of the directory with the passed
    dirname to be importable modules and packages (respectively) for the
    remainder of the active Python interpreter if this directory has yet to be
    registered *or* reduce to a noop otherwise (i.e., if this directory is
    registered already).

    Specifically, this function appends this dirname to the current
    :data:`sys.path` listing (in order) the dirnames of all directories to be
    iteratively searched for any module or package on first importing that
    module or package. To comply with Python standards in which the first item
    of this list is either the dirname of the directory containing the script
    from which this process was invoked *or* the empty string (signifying the
    current directory), this list is appended to rather than prepended to.

    Parameters
    ----------
    dirname : str
        Absolute or relative path of the directory to be registered.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs

    # Log this addition.
    logs.log_debug('Registering import directory: %s', dirname)

    # If this directory does *NOT* exist or is unreadable, raise an exception.
    dirs.die_unless_dir(dirname)

    # If the current PYTHONPATH already contains this directory...
    if dirname in sys.path:
        # Log this edge case.
        logs.log_debug('Ignoring already registered import directory.')

        # Reduce to a noop.
        return
    # Else, the current PYTHONPATH does *NOT* already contain this directory.

    # Append this directory to the current PYTHONPATH.
    sys.path.append(dirname)
